# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import itertools

import numpy as np
import pytest

import sisl
from sisl.viz.plotters.matrix import draw_matrix_separators, set_matrix_axes


def test_draw_matrix_separators_empty():
    C = sisl.Atom(
        "C",
        orbitals=[
            sisl.AtomicOrbital("2s"),
            sisl.AtomicOrbital("2px"),
            sisl.AtomicOrbital("2py"),
            sisl.AtomicOrbital("2pz"),
        ],
    )
    geom = sisl.geom.graphene(atoms=C)

    # Check combinations that should give no lines
    assert draw_matrix_separators(False, geom, "orbitals", "orbitals") == []
    assert draw_matrix_separators(True, geom, "atoms", "orbitals") == []


@pytest.mark.parametrize(
    "draw_supercells,separator_mode",
    itertools.product([True, False], ["atoms", "orbitals", "supercells"]),
)
def test_draw_matrix_separators(draw_supercells, separator_mode):
    C = sisl.Atom(
        "C",
        orbitals=[
            sisl.AtomicOrbital("2s"),
            sisl.AtomicOrbital("2px"),
            sisl.AtomicOrbital("2py"),
            sisl.AtomicOrbital("2pz"),
        ],
    )
    geom = sisl.geom.graphene(atoms=C)

    lines = draw_matrix_separators(
        {"color": "red"},
        geom,
        "orbitals",
        separator_mode=separator_mode,
        draw_supercells=draw_supercells,
    )

    if not draw_supercells and separator_mode == "supercells":
        assert len(lines) == 0
        return

    assert len(lines) == 1
    assert isinstance(lines[0], dict)
    action = lines[0]
    assert action["method"] == "draw_line"
    # Check that the number of points in the line is fine
    n_expected_points = {
        ("atoms", False): 6,
        ("atoms", True): 30,
        ("orbitals", False): 12,
        ("orbitals", True): 60,
        ("supercells", True): 24,
    }[separator_mode, draw_supercells]

    assert action["kwargs"]["x"].shape == (n_expected_points,)
    assert action["kwargs"]["y"].shape == (n_expected_points,)

    assert action["kwargs"]["line"]["color"] == "red"


def test_set_matrix_axes():

    C = sisl.Atom(
        "C",
        orbitals=[
            sisl.AtomicOrbital("2s"),
            sisl.AtomicOrbital("2px"),
            sisl.AtomicOrbital("2py"),
            sisl.AtomicOrbital("2pz"),
        ],
    )
    geom = sisl.geom.graphene(atoms=C)

    matrix = np.zeros((geom.no, geom.no * geom.n_s))

    actions = set_matrix_axes(
        matrix, geom, "orbitals", constrain_axes=False, set_labels=False
    )
    assert len(actions) == 1
    assert actions[0]["method"] == "set_axes_equal"

    # Test without labels
    actions = set_matrix_axes(
        matrix, geom, "orbitals", constrain_axes=True, set_labels=False
    )
    assert len(actions) == 3
    assert actions[0]["method"] == "set_axes_equal"
    assert actions[1]["method"] == "set_axis"
    assert actions[1]["kwargs"]["axis"] == "x"
    assert actions[1]["kwargs"]["range"] == [-0.5, geom.no * geom.n_s - 0.5]
    assert "tickvals" not in actions[1]["kwargs"]
    assert "ticktext" not in actions[1]["kwargs"]

    assert actions[2]["method"] == "set_axis"
    assert actions[2]["kwargs"]["axis"] == "y"
    assert actions[2]["kwargs"]["range"] == [geom.no - 0.5, -0.5]
    assert "tickvals" not in actions[2]["kwargs"]
    assert "ticktext" not in actions[2]["kwargs"]

    # Test with labels
    actions = set_matrix_axes(
        matrix, geom, "orbitals", constrain_axes=True, set_labels=True
    )
    assert len(actions) == 3
    assert actions[0]["method"] == "set_axes_equal"
    assert actions[1]["method"] == "set_axis"
    assert actions[1]["kwargs"]["axis"] == "x"
    assert actions[1]["kwargs"]["range"] == [-0.5, geom.no * geom.n_s - 0.5]
    assert np.all(actions[1]["kwargs"]["tickvals"] == np.arange(geom.no * geom.n_s))
    assert len(actions[1]["kwargs"]["ticktext"]) == geom.no * geom.n_s

    assert actions[2]["method"] == "set_axis"
    assert actions[2]["kwargs"]["axis"] == "y"
    assert actions[2]["kwargs"]["range"] == [geom.no - 0.5, -0.5]
    assert np.all(actions[2]["kwargs"]["tickvals"] == np.arange(geom.no))
    assert len(actions[2]["kwargs"]["ticktext"]) == geom.no



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import units
from sisl.io.wannier90 import *

pytestmark = [pytest.mark.io, pytest.mark.wannier90, pytest.mark.w90]


@pytest.mark.parametrize("unit", ["", "Ang\n", "ang\n", "bohr\n"])
def test_seedname_read_frac(sisl_tmp, unit):
    f = sisl_tmp("read_frac.win")
    with open(f, "w") as fh:
        fh.write(
            """
begin unit_cell_cart
  {} 2. 0. 0.
  0. 2. 0
  0 0 2.
end unit_cell_cart

begin atoms_frac
  C 0.5 0.5 0.5
end
""".format(
                unit
            )
        )
    g = winSileWannier90(f).read_geometry(order=["win"])

    if len(unit) == 0:
        unit = "ang"
    unit = units(unit.strip().capitalize(), "Ang")

    assert np.allclose(g.cell, np.identity(3) * 2 * unit)
    assert np.allclose(g.xyz, [1 * unit] * 3)


@pytest.mark.parametrize("unit_sc", ["", "Ang\n", "ang\n", "bohr\n"])
@pytest.mark.parametrize("unit", ["", "Ang\n", "ang\n", "bohr\n"])
def test_seedname_read_coord(sisl_tmp, unit_sc, unit):
    f = sisl_tmp("read_coord.win")
    with open(f, "w") as fh:
        fh.write(
            """
begin unit_cell_cart
  {} 2. 0. 0.
  0. 2. 0
  0 0 2.
end unit_cell_cart

begin atoms_cart
  {} C 0.5 0.5 0.5
end
""".format(
                unit_sc, unit
            )
        )
    g = winSileWannier90(f).read_geometry(order=["win"])

    if len(unit) == 0:
        unit = "ang"
    unit = units(unit.strip().capitalize(), "Ang")

    if len(unit_sc) == 0:
        unit_sc = "ang"
    unit_sc = units(unit_sc.strip().capitalize(), "Ang")

    assert np.allclose(g.cell, np.identity(3) * 2 * unit_sc)
    assert np.allclose(g.xyz, [0.5 * unit] * 3)


@pytest.mark.parametrize("frac", [True, False])
def test_seedname_write_read(sisl_tmp, sisl_system, frac):
    f = sisl_tmp("write_read.win")
    sile = winSileWannier90(f, "w")
    sile.write_geometry(sisl_system.g, frac=frac)

    g = winSileWannier90(f).read_geometry(order=["win"])
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)


def test_seedname_read_ham(sisl_files):
    f = winSileWannier90(sisl_files("wannier90", "silicon", "silicon.win"))

    ham = {}
    for key in ["hr", "tb"]:
        ham[key] = f.read_hamiltonian(cutoff=1e-4, order=[key])
        if not key == "hr":
            assert ham["hr"].spsame(ham[key])


def test_seedname_read_lattice(sisl_files):
    f = winSileWannier90(sisl_files("wannier90", "silicon", "silicon.win"))

    lat1 = f.read_lattice(order="tb")
    lat2 = f.read_lattice(order="win")
    assert np.allclose(lat1.cell, lat2.cell)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom
from sisl.io.fhiaims._geometry import *

pytestmark = [pytest.mark.io, pytest.mark.fhiaims, pytest.mark.aims]


def test_in_simple(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.in")
    sisl_system.g.write(inSileFHIaims(f, "w"))
    g = inSileFHIaims(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    assert sisl_system.g.atoms.equal(g.atoms, R=False)


def test_in_velocity(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.in")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    v = np.random.rand(len(g), 3)
    g.write(inSileFHIaims(f, "w"), velocity=v)

    # Try to read in different ways
    g2 = inSileFHIaims(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)

    g2, v2 = inSileFHIaims(f).read_geometry(ret_velocity=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(v, v2)

    # Try to read in different ways
    v2 = inSileFHIaims(f).read_velocity()
    assert np.allclose(v, v2)


def test_in_moment(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.in")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    m = np.random.rand(len(g))
    g.write(inSileFHIaims(f, "w"), moment=m)

    # Try to read in different ways
    g2 = inSileFHIaims(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)

    g2, m2 = inSileFHIaims(f).read_geometry(ret_moment=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(m, m2)

    # Try to read in different ways
    m2 = inSileFHIaims(f).read_moment()
    assert np.allclose(m, m2)


def test_in_v_m(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.in")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    v = np.random.rand(len(g), 3)
    m = np.random.rand(len(g))
    g.write(inSileFHIaims(f, "w"), velocity=v, moment=m)

    # Try to read in different ways
    g2 = inSileFHIaims(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)

    g2, m2 = inSileFHIaims(f).read_geometry(ret_moment=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(m, m2)

    g2, v2 = inSileFHIaims(f).read_geometry(ret_velocity=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(v, v2)

    g2, v2, m2 = inSileFHIaims(f).read_geometry(ret_velocity=True, ret_moment=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(v, v2)
    assert np.allclose(m, m2)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.vasp.locpot import *

pytestmark = [pytest.mark.io, pytest.mark.vasp]


def test_graphene_locpot(sisl_files):
    f = sisl_files("vasp", "graphene", "LOCPOT")
    gridf64 = locpotSileVASP(f).read_grid()
    gridf32 = locpotSileVASP(f).read_grid(dtype=np.float32)
    geom = locpotSileVASP(f).read_geometry()

    assert gridf64.dtype == np.float64
    assert gridf32.dtype == np.float32
    assert geom == gridf32.geometry

    gridHa = locpotSileVASP(f).read_grid(units="Ha")
    assert not np.allclose(gridf64.grid, gridHa.grid)


def test_graphene_locpot_index_float(sisl_files):
    f = sisl_files("vasp", "graphene", "LOCPOT")
    grid = locpotSileVASP(f).read_grid()
    gridh = locpotSileVASP(f).read_grid(index=[0.5])

    assert grid.grid.sum() / 2 == pytest.approx(gridh.grid.sum())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np

if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
    from numpy.exceptions import ComplexWarning
else:
    from numpy import ComplexWarning

import pytest

from sisl import Atom, DynamicalMatrix, Geometry, Lattice

pytestmark = [pytest.mark.physics, pytest.mark.dynamicalmatrix]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )

            C = Atom(Z=6, R=[bond * 1.01] * 3)
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )
            self.D = DynamicalMatrix(self.g)

            def func(D, ia, idxs, idxs_xyz):
                idx = D.geometry.close(
                    ia, R=(0.1, 1.44), atoms=idxs, atoms_xyz=idxs_xyz
                )
                ia = ia * 3

                i0 = idx[0] * 3
                i1 = idx[1] * 3
                # on-site
                p = 1.0
                D.D[ia, i0] = p
                D.D[ia + 1, i0 + 1] = p
                D.D[ia + 2, i0 + 2] = p

                # nn
                p = 0.1

                # on-site directions
                D.D[ia, ia + 1] = p
                D.D[ia, ia + 2] = p
                D.D[ia + 1, ia] = p
                D.D[ia + 1, ia + 2] = p
                D.D[ia + 2, ia] = p
                D.D[ia + 2, ia + 1] = p

                D.D[ia, i1 + 1] = p
                D.D[ia, i1 + 2] = p

                D.D[ia + 1, i1] = p
                D.D[ia + 1, i1 + 2] = p

                D.D[ia + 2, i1] = p
                D.D[ia + 2, i1 + 1] = p

            self.func = func

    return t()


class TestDynamicalMatrix:
    def test_objects(self, setup):
        assert len(setup.D.xyz) == 2
        assert setup.g.no == len(setup.D)

    def test_dtype(self, setup):
        assert setup.D.dtype == np.float64

    def test_ortho(self, setup):
        assert setup.D.orthogonal

    def test_set1(self, setup):
        setup.D.D[0, 0] = 1.0
        assert setup.D[0, 0] == 1.0
        assert setup.D[1, 0] == 0.0
        setup.D.empty()

    def test_apply_newton(self, setup):
        setup.D.construct(setup.func)
        assert setup.D[0, 0] == 1.0
        assert setup.D[1, 0] == 0.1
        assert setup.D[0, 1] == 0.1
        setup.D.apply_newton()
        setup.D.empty()

    def test_eig(self, setup):
        D = setup.D.copy()
        D.construct(setup.func)
        ev = D.eigenvalue()
        em = D.eigenmode()
        assert np.allclose(ev.hw, em.hw)
        assert np.allclose(em.norm2(), 1)

    def test_change_gauge(self, setup):
        D = setup.D.copy()
        D.construct(setup.func)
        em = D.eigenmode(k=(0.2, 0.2, 0.2))
        em2 = em.copy()
        em2.change_gauge("orbital")
        assert not np.allclose(em.mode, em2.mode)
        em2.change_gauge("cell")
        assert np.allclose(em.mode, em2.mode)

    @pytest.mark.filterwarnings("ignore", category=ComplexWarning)
    def test_dos_pdos_velocity(self, setup):
        D = setup.D.copy()
        D.construct(setup.func)
        E = np.linspace(0, 0.5, 10)
        em = D.eigenmode()
        assert np.allclose(em.DOS(E), em.PDOS(E).sum(0))

    def test_displacement(self, setup):
        D = setup.D.copy()
        D.construct(setup.func)
        em = D.eigenmode()
        assert em.displacement().shape == (len(em), D.geometry.na, 3)

    def test_pickle(self, setup):
        import pickle as p

        D = setup.D.copy()
        D.construct(setup.func)
        s = p.dumps(D)
        d = p.loads(s)
        assert d.spsame(D)
        assert np.allclose(d.eigh(), D.eigh())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
from scipy.sparse import SparseEfficiencyWarning

import sisl
from sisl import (
    Atom,
    Bloch,
    BrillouinZone,
    Geometry,
    Hamiltonian,
    Lattice,
    RealSpaceSE,
    RealSpaceSI,
    RecursiveSI,
    SemiInfinite,
    WideBandSE,
)

pytestmark = [
    pytest.mark.physics,
    pytest.mark.self_energy,
    pytest.mark.filterwarnings("ignore", category=SparseEfficiencyWarning),
]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )

            C = Atom(Z=6, R=[bond * 1.01])
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )
            self.H = Hamiltonian(self.g)
            func = self.H.create_construct([0.1, bond + 0.1], [0.0, -2.7])
            self.H.construct(func)
            self.HS = Hamiltonian(self.g, orthogonal=False)
            func = self.HS.create_construct(
                [0.1, bond + 0.1], [(0.0, 1.0), (-2.7, 0.0)]
            )
            self.HS.construct(func)

    return t()


def test_objects(setup):
    for D, si, sid in [("+A", 0, 1), ("-A", 0, -1), ("+B", 1, 1), ("-B", 1, -1)]:
        SE = SemiInfinite(setup.H.copy(), D)
        assert SE.semi_inf == si
        assert SE.semi_inf_dir == sid
        assert D in str(SE)


def test_sancho_orthogonal(setup):
    SE = RecursiveSI(setup.H.copy(), "+A")
    assert not np.allclose(SE.self_energy(0.1), SE.self_energy(0.1, bulk=True))


def test_sancho_bloch_zero_off_diag(setup):
    # check whether the bloch-expansion with 0 transverse
    # connections returns a block-diagonal matrix
    H = setup.H.copy()
    # disconnect transverse directions
    H.set_nsc(b=1)
    no = len(H)
    SE = RecursiveSI(H, "+A")

    for nb in [2, 4, 5]:
        bloch = Bloch(1, nb, 1)
        for E in [-0.1, 0.2, 0.4, -0.5]:
            se = bloch(SE.self_energy, [0.2, 0.2, 0.2], E=E)
            for b in range(1, nb):
                off = b * no
                assert np.allclose(se[:no, :no], se[off : off + no, off : off + no])
                se[off : off + no, off : off + no] = 0.0
            se[:no, :no] = 0.0
            assert np.allclose(se, 0.0)


def test_sancho_orthogonal_dtype(setup):
    SE = RecursiveSI(setup.H, "+A")
    s64 = SE.self_energy(0.1, dtype=np.complex64)
    s128 = SE.self_energy(0.1)
    assert s64.dtype == np.complex64
    assert s128.dtype == np.complex128
    assert np.allclose(s64, s128)


def test_sancho_warning():
    lattice = Lattice([1, 1, 10], nsc=[5, 5, 1])
    C = Atom(Z=6, R=[2 * 1.01])
    g = Geometry([[0.0, 0.0, 0.0]], atoms=C, lattice=lattice)
    H = Hamiltonian(g)
    func = H.create_construct([0.1, 1.01, 2.01], [0.0, -2.0, -1.0])
    H.construct(func)

    with pytest.warns(sisl.SislWarning, match=r"first neighboring cell.*\[1.\]"):
        RecursiveSI(H, "+A")


def test_sancho_non_orthogonal(setup):
    SE = RecursiveSI(setup.HS, "-A")
    assert not np.allclose(SE.self_energy(0.1), SE.self_energy(0.1, bulk=True))


def test_sancho_broadening_matrix(setup):
    SE = RecursiveSI(setup.HS, "-A")
    assert np.allclose(SE.broadening_matrix(0.1), SE.se2broadening(SE.self_energy(0.1)))


def test_sancho_non_orthogonal_dtype(setup):
    SE = RecursiveSI(setup.HS, "-A")
    s64 = SE.self_energy(0.1, dtype=np.complex64)
    s128 = SE.self_energy(0.1)
    assert s64.dtype == np.complex64
    assert s128.dtype == np.complex128
    assert np.allclose(s64, s128)


def test_sancho_lr(setup):
    SL = RecursiveSI(setup.HS, "-A")
    SR = RecursiveSI(setup.HS, "+A")

    E = 0.1
    k = [0, 0.13, 0]

    # Check that left/right are different

    L_SE = SL.self_energy(E, k)
    R_SE = SR.self_energy(E, k)
    assert not np.allclose(L_SE, R_SE)

    LB_SEL, LB_SER = SL.self_energy_lr(E, k)
    RB_SEL, RB_SER = SR.self_energy_lr(E, k)

    assert np.allclose(LB_SEL, L_SE)
    assert np.allclose(LB_SER, R_SE)
    assert np.allclose(RB_SEL, L_SE)
    assert np.allclose(RB_SER, R_SE)

    LB_SEL, LB_SER = SL.self_energy_lr(E, k, bulk=True)
    L_SE = SL.self_energy(E, k, bulk=True)
    R_SE = SR.self_energy(E, k, bulk=True)
    assert not np.allclose(L_SE, R_SE)
    RB_SEL, RB_SER = SR.self_energy_lr(E, k, bulk=True)

    assert np.allclose(LB_SEL, L_SE)
    assert np.allclose(LB_SER, R_SE)
    assert np.allclose(RB_SEL, L_SE)
    assert np.allclose(RB_SER, R_SE)


def test_sancho_green(setup):
    SL = RecursiveSI(setup.HS, "-A")
    SR = RecursiveSI(setup.HS, "+A")

    E = 0.1
    k = [0, 0.13, 0]

    # Check that left/right are different

    L_SE = SL.self_energy(E, k, bulk=True)
    R_SE = SR.self_energy(E, k)
    g = np.linalg.inv(L_SE - R_SE)
    G = SL.green(E, k)
    assert np.allclose(g, G)
    assert np.allclose(SL.green(E, k), SR.green(E, k))


def test_wideband_1(setup):
    SE = WideBandSE(10, 1e-2)
    se = SE.self_energy()
    assert se.shape == (10, 10)
    assert np.allclose(np.diag(se), -1j * np.pi * 1e-2)
    assert np.allclose(np.diag(SE.self_energy(1j)), -1j * np.pi * 1.0)
    assert np.allclose(np.diag(SE.broadening_matrix(1j)), np.pi * 2.0)
    # ensure our custom function works!
    assert np.allclose(SE.broadening_matrix(1j), SE.se2broadening(SE.self_energy(1j)))


@pytest.mark.parametrize("k_axes", [0, 1])
@pytest.mark.parametrize("semi_axis", [0, 1])
@pytest.mark.parametrize("trs", [True, False])
@pytest.mark.parametrize("bz", [None, BrillouinZone([1])])
@pytest.mark.parametrize("unfold", [1, 2])
def test_real_space_HS(setup, k_axes, semi_axis, trs, bz, unfold):
    if k_axes == semi_axis:
        return
    RSE = RealSpaceSE(setup.HS, semi_axis, k_axes, (unfold, unfold, unfold))
    RSE.set_options(dk=100, trs=trs, bz=bz)
    RSE.setup()
    RSE.green(0.1)


@pytest.mark.parametrize("k_axes", [0, 1])
@pytest.mark.parametrize("semi_axis", [0, 1])
@pytest.mark.parametrize("trs", [True, False])
@pytest.mark.parametrize("bz", [None, BrillouinZone([1])])
@pytest.mark.parametrize("unfold", [1, 2])
def test_real_space_H(setup, k_axes, semi_axis, trs, bz, unfold):
    if k_axes == semi_axis:
        return
    RSE = RealSpaceSE(
        setup.H, semi_axis, k_axes, (unfold, unfold, 1), trs=trs, dk=100, bz=bz
    )
    RSE.green(0.1)
    RSE.self_energy(0.1)


def test_real_space_H_3d():
    lattice = Lattice(1.0, nsc=[3] * 3)
    H = Atom(Z=1, R=[1.001])
    geom = Geometry([0] * 3, atoms=H, lattice=lattice)
    H = Hamiltonian(geom)
    H.construct(([0.001, 1.01], (0, -1)))
    RSE = RealSpaceSE(H, 0, [1, 2], (3, 4, 2))
    RSE.set_options(dk=100, trs=True)
    RSE.setup()
    nk = np.ones(3, np.int32)
    nk[[1, 2]] = 23
    bz = BrillouinZone(H, nk)
    RSE.set_options(bz=bz)

    RSE.green(0.1)
    # Since there is only 2 repetitions along one direction we will have the full matrix
    # coupled!
    assert np.allclose(RSE.self_energy(0.1), RSE.self_energy(0.1, coupling=True))
    assert np.allclose(
        RSE.self_energy(0.1, bulk=True), RSE.self_energy(0.1, bulk=True, coupling=True)
    )


@pytest.mark.xfail(
    int(np.__version__.split(".")[0]) >= 2,
    reason="some numpy2 bug means it will not honor dtype=complex64",
)
def test_real_space_H_dtype(setup):
    RSE = RealSpaceSE(setup.H, 0, 1, (2, 2, 1), dk=100)
    g64 = RSE.green(0.1, dtype=np.complex64)
    g128 = RSE.green(0.1, dtype=np.complex128)
    assert g64.dtype == np.complex64
    assert g128.dtype == np.complex128
    assert np.allclose(g64, g128, atol=1.0e-4)

    s64 = RSE.self_energy(0.1, dtype=np.complex64)
    s128 = RSE.self_energy(0.1, dtype=np.complex128)
    assert s64.dtype == np.complex64
    assert s128.dtype == np.complex128
    assert np.allclose(s64, s128, atol=1e-2)
    RSE.real_space_coupling()
    RSE.clear()


def test_real_space_H_SE_unfold(setup):
    # check that calculating the real-space Green function is equivalent for two equivalent systems
    RSE = RealSpaceSE(setup.H, 0, 1, (2, 2, 1), dk=100)
    RSE_big = RealSpaceSE(setup.H.tile(2, 0).tile(2, 1), semi_axis=0, k_axes=1, dk=100)

    for E in [0.1, 1.5]:
        G = RSE.green(E)
        G_big = RSE_big.green(E)
        assert np.allclose(G, G_big)

        SE = RSE.self_energy(E)
        SE_big = RSE_big.self_energy(E)
        assert np.allclose(SE, SE_big)


def test_real_space_HS_SE_unfold(setup):
    # check that calculating the real-space Green function is equivalent for two equivalent systems
    RSE = RealSpaceSE(setup.HS, 0, 1, (2, 2, 1), dk=100)
    RSE_big = RealSpaceSE(setup.HS.tile(2, 0).tile(2, 1), semi_axis=0, k_axes=1, dk=100)

    for E in [0.1, 1.5]:
        G = RSE.green(E)
        G_big = RSE_big.green(E)
        assert np.allclose(G, G_big)

        SE = RSE.self_energy(E)
        SE_big = RSE_big.self_energy(E)
        assert np.allclose(SE, SE_big)


def test_real_space_HS_SE_unfold_with_k():
    # check that calculating the real-space Green function is equivalent for two equivalent systems
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3] * 3)
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    RSE = RealSpaceSE(H, 0, 1, (3, 4, 1), dk=100, trs=False)

    k1 = [0, 0, 0.2]
    k2 = [0, 0, 0.3]
    for E in [0.1, 1.5]:
        G1 = RSE.green(E, k1)
        G2 = RSE.green(E, k2)
        assert not np.allclose(G1, G2)

        SE1 = RSE.self_energy(E, k1)
        SE2 = RSE.self_energy(E, k2)
        assert not np.allclose(SE1, SE2)


def test_real_space_SE_fail_k_trs():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3] * 3)
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    RSE = RealSpaceSE(H, 0, 1, (3, 4, 1))
    with pytest.raises(ValueError):
        RSE.green(0.1, [0, 0, 0.2])


def test_real_space_SE_fail_k_semi_same():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3] * 3)
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    with pytest.raises(ValueError):
        RSE = RealSpaceSE(H, 0, 0, (3, 4, 1))


def test_real_space_SE_fail_nsc_semi():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([5, 3, 1])
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    with pytest.raises(ValueError):
        RSE = RealSpaceSE(H, 0, 1, (3, 4, 1), dk=100)


def test_real_space_SE_fail_nsc_k():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3, 1, 1])
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    with pytest.raises(ValueError):
        RSE = RealSpaceSE(H, 0, 1, (3, 4, 1), dk=100)


def test_real_space_SE_fail_nsc_semi_fully_periodic():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3, 5, 3])
    H = Hamiltonian(sq)
    H.construct([(0.1, 1.1), (4, -1)])

    with pytest.raises(ValueError):
        RSE = RealSpaceSE(H, 1, 0, (3, 4, 1), dk=100)


def test_real_space_SE_spin_orbit():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3, 3, 3])
    H = Hamiltonian(sq, spin="SOC")
    on = [4, 0, 0, 0, 0, 0, 0, 0]
    off = [-1, 0, 0, 0, 0, 0, 0, 0]
    H.construct([(0.1, 1.1), (on, off)])
    RS = RealSpaceSE(H, 0, (1, 2), (2, 2, 2), dk=1.4)
    RS.self_energy(0.1)
    RS.self_energy(0.1, coupling=True)


@pytest.mark.parametrize("k_axes", [0])
@pytest.mark.parametrize("trs", [True, False])
@pytest.mark.parametrize("bz", [None, BrillouinZone([1])])
@pytest.mark.parametrize("unfold", [1, 3])
@pytest.mark.parametrize("bulk", [True, False])
@pytest.mark.parametrize("coupling", [True, False])
def test_real_space_SI_HS(setup, k_axes, trs, bz, unfold, bulk, coupling):
    semi = RecursiveSI(setup.HS, "-B")
    surf = setup.HS.tile(4, 1)
    surf.set_nsc(b=1)
    RSI = RealSpaceSI(semi, surf, k_axes, (unfold, 1, unfold))
    RSI.setup(dk=100, trs=trs, bz=bz)
    RSI.self_energy(0.1, bulk=bulk, coupling=coupling)


@pytest.mark.parametrize("semi_dir", ["-B", "+B"])
@pytest.mark.parametrize("k_axes", [0])
@pytest.mark.parametrize("trs", [True, False])
@pytest.mark.parametrize("bz", [None, BrillouinZone([1])])
@pytest.mark.parametrize("unfold", [1, 3])
@pytest.mark.parametrize("bulk", [True, False])
@pytest.mark.parametrize("semi_bulk", [True, False])
@pytest.mark.parametrize("coupling", [True, False])
def test_real_space_SI_H(
    setup, semi_dir, k_axes, trs, bz, unfold, bulk, semi_bulk, coupling
):
    semi = RecursiveSI(setup.H, semi_dir)
    surf = setup.H.tile(4, 1)
    surf.set_nsc(b=1)
    RSI = RealSpaceSI(semi, surf, k_axes, (unfold, 1, unfold))
    RSI.set_options(dk=100, trs=trs, bz=bz, semi_bulk=semi_bulk)
    RSI.setup()
    RSI.self_energy(0.1, bulk=bulk, coupling=coupling)


def test_real_space_SI_H_test(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    surf.set_nsc(b=1)
    RSI = RealSpaceSI(semi, surf, 0, (3, 1, 3))
    RSI.set_options(dk=100, trs=False, bz=None)
    RSI.setup()
    RSI.green(0.1, [0, 0, 0.1])
    RSI.self_energy(0.1, [0, 0, 0.1])
    RSI.clear()


def test_real_space_SI_H_k_trs(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    surf.set_nsc(b=1)
    RSI = RealSpaceSI(semi, surf, 0, (3, 1, 3))
    RSI.setup(dk=100, trs=True, bz=None)
    with pytest.raises(ValueError):
        RSI.green(0.1, [0, 0, 0.1])


def test_real_space_SI_fail_semi_in_k(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    surf.set_nsc(b=1)
    with pytest.raises(ValueError):
        RSI = RealSpaceSI(semi, surf, [0, 1], (2, 1, 1))


def test_real_space_SI_fail_surf_nsc(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    with pytest.raises(ValueError):
        RSI = RealSpaceSI(semi, surf, 0, (2, 1, 1))


def test_real_space_SI_fail_k_no_nsc(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    surf.set_nsc([1] * 3)
    with pytest.raises(ValueError):
        RSI = RealSpaceSI(semi, surf, 0, (2, 1, 1))


def test_real_space_SI_fail_unfold_in_semi(setup):
    semi = RecursiveSI(setup.H, "-B")
    surf = setup.H.tile(4, 1)
    surf.set_nsc(b=1)
    with pytest.raises(ValueError):
        RSI = RealSpaceSI(semi, surf, 0, (2, 2, 1))


def test_real_space_SI_spin_orbit():
    sq = Geometry([0] * 3, Atom(1, 1.01), [1])
    sq.set_nsc([3, 3, 1])
    H = Hamiltonian(sq, spin="SOC")
    on = [4, 0, 0, 0, 0, 0, 0, 0]
    off = [-1, 0, 0, 0, 0, 0, 0, 0]
    H.construct([(0.1, 1.1), (on, off)])
    semi = RecursiveSI(H, "-B")
    surf = H.tile(4, 1)
    surf.set_nsc(b=1)
    RS = RealSpaceSI(semi, surf, 0, (2, 1, 1), dk=1.5)
    RS.self_energy(0.1)
    RS.self_energy(0.1, coupling=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import warnings

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def si_pdos_kgrid_geom(with_orbs=True):
    if with_orbs:
        return sisl.geom.diamond(5.43, sisl.Atom("Si", R=np.arange(13) + 1))
    return sisl.geom.diamond(5.43, sisl.Atom("Si"))


@pytest.mark.parametrize("order", ["HSX", "TSHS"])
def test_si_pdos_kgrid_H_eig(sisl_files, sisl_tmp, order):
    fdf = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    H = fdf.read_hamiltonian(order=order)

    eigs = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.EIG")).read_data()
    bz = sisl.get_sile(
        sisl_files("siesta", "Si_pdos_k", "Si_pdos.KP")
    ).read_brillouinzone(H.geometry)

    # Now see if we get the same data...
    for ispin in range(eigs.shape[0]):
        for i, (eig, k) in enumerate(zip(eigs[ispin], bz)):
            if i > 10:
                break
            Heig = H.eigh(k, spin=ispin)
            assert np.allclose(eig, Heig)


def test_si_pdos_kgrid_hsx_H(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.HSX"))
    si.read_hamiltonian(geometry=si_pdos_kgrid_geom())


def test_si_pdos_kgrid_hsx_H(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    si.read_hamiltonian(order="HSX")


def test_si_pdos_kgrid_hsx_H_no_geometry(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.HSX"))
    H0 = si.read_hamiltonian()
    H1 = si.read_hamiltonian(geometry=si_pdos_kgrid_geom())
    assert H0._csr.spsame(H1._csr)


def test_si_pdos_kgrid_hsx_H_fix_orbitals(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.HSX"))
    si.read_hamiltonian(geometry=si_pdos_kgrid_geom(False))


def test_si_pdos_kgrid_hsx_overlap(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.HSX"))
    HS = si.read_hamiltonian(geometry=si_pdos_kgrid_geom())
    S = si.read_overlap(geometry=si_pdos_kgrid_geom(False))

    assert HS._csr.spsame(S._csr)
    assert np.allclose(HS._csr._D[:, HS.S_idx], S._csr._D[:, 0])


def test_h2o_dipole_hsx_no_geometry(sisl_files, sisl_tmp):
    HSX = sisl.get_sile(sisl_files("siesta", "H2O_dipole", "h2o_dipole.HSX"))
    geometry = sisl.get_sile(
        sisl_files("siesta", "H2O_dipole", "h2o_dipole.fdf")
    ).read_geometry()
    # manually define this.
    geometry.set_nsc(a=5, b=1, c=3)
    HS = HSX.read_hamiltonian()
    S = HSX.read_overlap(geometry=geometry)

    assert HS._csr.spsame(S._csr)
    assert np.allclose(HS._csr._D[:, HS.S_idx], S._csr._D[:, 0])


def test_h2o_dipole_hsx(sisl_files, sisl_tmp):
    HSX = sisl.get_sile(sisl_files("siesta", "H2O_dipole", "h2o_dipole.HSX"))
    geometry = sisl.get_sile(
        sisl_files("siesta", "H2O_dipole", "h2o_dipole.fdf")
    ).read_geometry()
    geometry.set_nsc(a=5, b=1, c=3)
    # reading from hsx just requires atoms + coordinates + nsc
    HS = HSX.read_hamiltonian(geometry=geometry)
    S = HSX.read_overlap(geometry=geometry)

    assert HS._csr.spsame(S._csr)
    assert np.allclose(HS._csr._D[:, HS.S_idx], S._csr._D[:, 0])


def test_h2o_dipole_hsx_hs_no_geometry(sisl_files, sisl_tmp):
    HSX = sisl.get_sile(sisl_files("siesta", "H2O_dipole", "h2o_dipole.HSX"))
    HS = HSX.read_hamiltonian()
    S = HSX.read_overlap()


def test_h2o_dipole_hsx_no_ef(sisl_files, sisl_tmp):
    HSX = sisl.get_sile(sisl_files("siesta", "H2O_dipole", "h2o_dipole.HSX"))
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        Ef = HSX.read_fermi_level()


def test_si_pdos_kgrid_hsx_versions(sisl_files, sisl_tmp):
    HSX0 = sisl.get_sile(sisl_files("siesta", "ancient", "si_pdos_kgrid.0.HSX"))
    assert HSX0.version == 0
    HSX1 = sisl.get_sile(sisl_files("siesta", "ancient", "si_pdos_kgrid.1.HSX"))
    assert HSX1.version == 1

    HS0 = HSX0.read_hamiltonian()
    HS1 = HSX1.read_hamiltonian()
    Ef = HSX1.read_fermi_level()
    # HSX0 does not shift, whereas HSX1 does shift
    HS1.shift(Ef)
    assert HS0._csr.spsame(HS1._csr)
    # assert np.allclose(HS0._csr._D, HS1._csr._D)


def test_si_pdos_kgrid_hsx_versions_s(sisl_files, sisl_tmp):
    HSX0 = sisl.get_sile(sisl_files("siesta", "ancient", "si_pdos_kgrid.0.HSX"))
    assert HSX0.version == 0
    HSX1 = sisl.get_sile(sisl_files("siesta", "ancient", "si_pdos_kgrid.1.HSX"))
    assert HSX1.version == 1

    HS0 = HSX0.read_hamiltonian()
    S0 = HSX0.read_overlap()
    HS1 = HSX1.read_hamiltonian()
    S1 = HSX1.read_overlap()
    assert HS0._csr.spsame(S1._csr)
    assert S0._csr.spsame(HS1._csr)

    assert np.allclose(HS0._csr._D[:, HS0.S_idx], S0._csr._D[:, 0])
    assert np.allclose(HS1._csr._D[:, HS1.S_idx], S1._csr._D[:, 0])


def test_si_pdos_kgrid_hsx_1_same_tshs(sisl_files, sisl_tmp):
    HSX = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.HSX"))
    TSHS = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))

    HSX = HSX.read_hamiltonian()
    TSHS = TSHS.read_hamiltonian()

    gx = HSX.geometry
    gt = TSHS.geometry
    assert np.allclose(gx.lattice.cell, gt.lattice.cell)
    assert np.allclose(gx.xyz, gt.xyz)
    assert np.allclose(gx.nsc, gt.nsc)


def test_hsx_si_pdos_kgrid_tofromnc(sisl_files, sisl_tmp):
    pytest.importorskip("netCDF4")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS1 = si.read_hamiltonian()
    f = sisl_tmp("tmp.HSX")
    fnc = sisl_tmp("tmp.nc")

    HS1.write(f)
    HS1.write(fnc)

    HS2 = sisl.get_sile(f).read_hamiltonian()
    HS2nc = sisl.get_sile(fnc).read_hamiltonian()
    assert HS1._csr.spsame(HS2._csr)
    assert HS1._csr.spsame(HS2nc._csr)
    HS1.finalize()
    HS2.finalize()
    assert np.allclose(HS1._csr._D, HS2._csr._D)
    HS2nc.finalize()
    assert np.allclose(HS1._csr._D, HS2nc._csr._D)


@pytest.mark.parametrize("version", [1, 2])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_hsx_si_pdos_kgrid_dtype(sisl_files, sisl_tmp, version, dtype):
    pytest.importorskip("netCDF4")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS = si.read_hamiltonian()
    f = sisl_tmp("tmp.1.HSX")

    HS.write(f, version=version, dtype=dtype)

    HS1 = sisl.get_sile(f).read_hamiltonian()
    assert HS1.dtype == dtype
    assert HS._csr.spsame(HS1._csr)
    HS.finalize()
    HS1.finalize()
    assert np.allclose(HS._csr._D, HS1._csr._D)


@pytest.mark.parametrize("order", ["HSX", "TSHS"])
def test_srti03_H_eig(sisl_files, sisl_tmp, order):
    fdf = sisl.get_sile(sisl_files("siesta", "SrTiO3", "non-collinear", "SrTiO3.fdf"))
    H = fdf.read_hamiltonian(order=order)

    eigs = sisl.get_sile(
        sisl_files("siesta", "SrTiO3", "non-collinear", "SrTiO3.EIG")
    ).read_data()
    bz = sisl.get_sile(
        sisl_files("siesta", "SrTiO3", "non-collinear", "SrTiO3.KP")
    ).read_brillouinzone(H.geometry)

    # Now see if we get the same data...
    for i, (eig, k) in enumerate(zip(eigs[0], bz)):
        if i > 10:
            break
        Heig = H.eigh(k)
        assert np.allclose(eig, Heig)


@pytest.mark.parametrize("order", ["HSX"])
def test_bi_hexagonal_H_eig(sisl_files, sisl_tmp, order):
    N = 20
    fdf = sisl.get_sile(sisl_files("siesta", "Bi_hexagonal", "Bi_hexagonal.fdf"))
    H = fdf.read_hamiltonian(order=order)

    eigs = sisl.get_sile(
        sisl_files("siesta", "Bi_hexagonal", "Bi_hexagonal.EIG")
    ).read_data()
    bz = sisl.get_sile(
        sisl_files("siesta", "Bi_hexagonal", "Bi_hexagonal.KP")
    ).read_brillouinzone(H.geometry)

    # Now see if we get the same data...
    for i, (eig, k) in enumerate(zip(eigs[0], bz)):
        if i > 10:
            break
        Heig = H.eigh(k)
        assert np.allclose(eig, Heig)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl
from sisl.viz.plots import geometry_plot

pytestmark = [pytest.mark.viz, pytest.mark.plots]


@pytest.fixture(scope="module", params=["plotly", "matplotlib"])
def backend(request):
    pytest.importorskip(request.param)
    return request.param


@pytest.fixture(scope="module", params=["x", "xy", "xyz"])
def axes(request):
    return request.param


@pytest.fixture(scope="module")
def geometry():
    return sisl.geom.graphene()


def test_geometry_plot(geometry, axes, backend):
    if axes == "xyz" and backend == "matplotlib":
        with pytest.raises(NotImplementedError):
            geometry_plot(geometry, axes=axes, backend=backend)
    else:
        geometry_plot(geometry, axes=axes, backend=backend)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl.viz.processors.logic import swap

pytestmark = [pytest.mark.viz, pytest.mark.processors]


def test_swap():
    assert swap(1, (1, 2)) == 2
    assert swap(2, (1, 2)) == 1

    with pytest.raises(ValueError):
        swap(3, (1, 2))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import warnings as w

import pytest

import sisl.messages as sm

pytestmark = pytest.mark.messages


def test_deprecate():
    with pytest.warns(sm.SislDeprecation):
        sm.deprecate("Deprecation warning")


def test_deprecation():
    with pytest.warns(sm.SislDeprecation):
        w.warn(sm.SislDeprecation("Deprecation warning"))


def test_warn_method():
    with pytest.warns(sm.SislWarning):
        sm.warn("Warning")


def test_warn_specific():
    with pytest.warns(sm.SislWarning):
        sm.warn(sm.SislWarning("Warning"))


def test_warn_category():
    with pytest.warns(sm.SislWarning):
        sm.warn("Warning", sm.SislWarning)


def test_info_method():
    with pytest.warns(sm.SislInfo):
        sm.info("Information")


def test_info_specific():
    with pytest.warns(sm.SislInfo):
        sm.info(sm.SislInfo("Info"))


def test_info_category():
    with pytest.warns(sm.SislInfo):
        sm.info("Information", sm.SislInfo)


def test_error():
    with pytest.raises(sm.SislError):
        raise sm.SislError("This is an error")


def test_exception():
    with pytest.raises(sm.SislException):
        raise sm.SislException("This is an error")


def test_progressbar_true():
    eta = sm.progressbar(2, "Hello", "unit", True)
    eta.update()
    eta.update()
    eta.close()


def test_progressbar_false():
    eta = sm.progressbar(2, "Hello", "unit", False)
    eta.update()
    eta.update()
    eta.close()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl.unit.codata import read_codata

pytestmark = pytest.mark.unit


@pytest.mark.parametrize("year", (2010, 2014, 2018, 2022))
def test_codata(year):
    codata = read_codata(year)
    assert codata["year"] == f"{year}"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" tests for sisl/linalg """



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.mixing import AndersonMixer, LinearMixer, StepMixer

pytestmark = pytest.mark.mixing


def test_step_manual():
    mix1 = LinearMixer()
    mix2 = AndersonMixer()

    # now merge them
    def gen():
        yield mix1
        yield mix1
        yield mix2
        yield mix2

    mixer = StepMixer(gen)
    # Now lets test it
    f = np.random.rand(4) + 1
    df = np.random.rand(4) + 0.2

    # test 3 times
    for _ in range(3):
        # we start with linearmixer
        for _ in range(2):
            assert isinstance(mixer.mixer, LinearMixer)
            mixer(f, df)

        for _ in range(2):
            assert isinstance(mixer.mixer, AndersonMixer)
            mixer(f, df)


def test_step_yield_repeat():
    mix1 = LinearMixer()
    mix2 = AndersonMixer()
    mix1_rep2 = StepMixer.yield_repeat(mix1, 2)
    mix2_rep2 = StepMixer.yield_repeat(mix2, 2)

    # now merge them
    def gen():
        yield from mix1_rep2()
        yield from mix2_rep2()

    mixer = StepMixer(gen)
    # Now lets test it
    f = np.random.rand(4) + 1
    df = np.random.rand(4) + 0.2

    # test 3 times
    for _ in range(3):
        # we start with linearmixer
        for _ in range(2):
            assert isinstance(mixer.mixer, LinearMixer)
            mixer(f, df)

        for _ in range(2):
            assert isinstance(mixer.mixer, AndersonMixer)
            mixer(f, df)


def test_step_yield_repeat_n():
    mix = LinearMixer()
    r = StepMixer.yield_repeat(mix, 1)

    g = r()
    next(g)
    with pytest.raises(StopIteration):
        next(g)

    r = StepMixer.yield_repeat(mix, 2)
    g = r()
    next(g)
    next(g)
    with pytest.raises(StopIteration):
        next(g)


def test_step_yield_chain():
    mix1 = LinearMixer()
    mix2 = AndersonMixer()
    mix1_rep2 = StepMixer.yield_repeat(mix1, 2)
    mix2_rep2 = StepMixer.yield_repeat(mix2, 2)

    # now merge them
    mixer = StepMixer(mix1_rep2, mix2_rep2)

    # Now lets test it
    f = np.random.rand(4) + 1
    df = np.random.rand(4) + 0.2

    # test 3 times
    for _ in range(3):
        # we start with linearmixer
        for _ in range(2):
            assert isinstance(mixer.mixer, LinearMixer)
            mixer(f, df)

        for _ in range(2):
            assert isinstance(mixer.mixer, AndersonMixer)
            mixer(f, df)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_wfsx_read(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Bi2Se3_3layer", "Bi2Se3.fdf"))
    wfsx = sisl.get_sile(
        sisl_files("siesta", "Bi2Se3_3layer", "Bi2Se3.bands.WFSX"),
        parent=fdf.read_geometry(),
    )

    info = wfsx.read_info()
    sizes = wfsx.read_sizes()
    basis = wfsx.read_basis()

    # yield states
    nstates = 0
    nstates_total = 0
    for state in wfsx.yield_eigenstate():
        nstates += 1
        nstates_total += len(state)
    # in this example we a SOC calculation, so we should not muliply by spin
    assert nstates == len(info[0])
    assert nstates_total == info[2].sum()

    bz = wfsx.read_brillouinzone()
    assert len(bz) == 16



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import os
import sys
from pathlib import Path

import numpy as np
import pytest

from sisl import (
    DensityMatrix,
    EnergyDensityMatrix,
    Geometry,
    Grid,
    Hamiltonian,
    Lattice,
)
from sisl._environ import sisl_environ
from sisl.io import *
from sisl.io.siesta.binaries import _gfSileSiesta
from sisl.io.tbtrans._cdf import *
from sisl.io.vasp import chgSileVASP
from sisl.io.wannier90 import winSileWannier90

pytestmark = pytest.mark.io


gs = get_sile
gsc = get_sile_class


def _my_intersect(a, b):
    a = get_siles(a)
    b = set(get_siles(b))
    return [k for k in a if k in b]


def _fnames(base, variants):
    files = [base + "." + v if len(v) > 0 else base for v in variants]
    files.append(Path(files[0]))
    return files


def test_get_sile1():
    cls = gsc("test.xyz")
    assert issubclass(cls, xyzSile)

    cls = gsc("test.regardless{xyz}")
    assert issubclass(cls, xyzSile)

    cls = gsc("test.fdf{xyz}")
    assert issubclass(cls, xyzSile)

    cls = gsc(Path("test.fdf{startswith=xyz}"))
    assert issubclass(cls, xyzSile)

    cls = gsc("test.xyz{fdf}")
    assert issubclass(cls, fdfSileSiesta)
    cls = gsc("test.cube{startswith=fdf}")
    assert issubclass(cls, fdfSileSiesta)


def test_get_sile2():
    # requesting non implemented files
    with pytest.raises(NotImplementedError):
        gsc("test.this_file_does_not_exist")


def test_get_out_context():
    with sisl_environ(SISL_IO_DEFAULT="siesta"):
        assert issubclass(gsc("test.out"), stdoutSileSiesta)
    with sisl_environ(SISL_IO_DEFAULT="orca"):
        assert issubclass(gsc("test.out"), stdoutSileORCA)


class TestObject:
    def test_siesta_sources(self):
        pytest.importorskip("sisl.io.siesta._siesta")

    @pytest.mark.parametrize("Sile", get_siles())
    def test_direct_string_instantiation(
        self,
        sisl_tmp,
        Sile,
    ):
        if issubclass(Sile, SileCDF):
            pytest.importorskip("netCDF4")

        # branch with string arguments
        fp = sisl_tmp("shouldnotexist.1234567")
        if os.path.exists(fp):
            os.remove(fp)

        if issubclass(Sile, SileCDF):
            sile = Sile(fp, _open=False)
        else:
            sile = Sile(fp)

        assert isinstance(sile, Sile)
        assert not os.path.exists(fp)

        # branch with Path arguments
        fp = Path(sisl_tmp("shouldnotexist.1234567"))
        if fp.exists():
            os.remove(str(fp))

        if issubclass(Sile, SileCDF):
            sile = Sile(fp, _open=False)
        else:
            sile = Sile(fp)

        assert isinstance(sile, Sile)
        assert not fp.exists()

    @pytest.mark.parametrize(
        "sile", _fnames("test", ["cube", "CUBE", "cube.gz", "CUBE.gz"])
    )
    def test_cube(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, cubeSile]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["ascii", "ascii.gz", "ascii.gz"]))
    def test_bigdft_ascii(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileBigDFT, asciiSileBigDFT]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["gout", "gout.gz"]))
    def test_gulp_gout(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileGULP, gotSileGULP]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["REF", "REF.gz"]))
    def test_scaleup_REF(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileScaleUp, refSileScaleUp]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["restart", "restart.gz"]))
    def test_scaleup_restart(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileScaleUp, restartSileScaleUp]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["rham", "rham.gz"]))
    def test_scaleup_rham(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileScaleUp, rhamSileScaleUp]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize(
        "sile", _fnames("test", ["fdf", "fdf.gz", "FDF.gz", "FDF"])
    )
    def test_siesta_fdf(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileSiesta, fdfSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["out", "out.gz"]))
    def test_siesta_out(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileSiesta, stdoutSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["nc"]))
    def test_siesta_nc(self, sile):
        pytest.importorskip("netCDF4")
        s = gs(sile, _open=False)
        for obj in [BaseSile, SileCDF, SileCDFSiesta, ncSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["grid.nc"]))
    def test_siesta_grid_nc(self, sile):
        pytest.importorskip("netCDF4")
        sile = gs(sile, _open=False)
        for obj in [BaseSile, SileCDF, SileCDFSiesta, gridncSileSiesta]:
            assert isinstance(sile, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["XV", "XV.gz"]))
    def test_siesta_xv(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileSiesta, xvSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["XV", "XV.gz"]))
    def test_siesta_xv_base(self, sile):
        s = gs(sile, cls=SileSiesta)
        for obj in [BaseSile, Sile, SileSiesta, xvSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["PDOS.xml", "PDOS.xml.gz"]))
    def test_siesta_pdos_xml(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileSiesta, pdosSileSiesta]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["ham", "HAM", "HAM.gz"]))
    def test_ham(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, hamiltonianSile]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["TBT.nc"]))
    def test_tbtrans_nc(self, sile):
        pytest.importorskip("netCDF4")
        s = gs(sile, _open=False)
        for obj in [BaseSile, SileCDF, SileCDFTBtrans, tbtncSileTBtrans]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["PHT.nc"]))
    def test_phtrans_nc(self, sile):
        pytest.importorskip("netCDF4")
        s = gs(sile, _open=False)
        for obj in [
            BaseSile,
            SileCDF,
            SileCDFTBtrans,
            tbtncSileTBtrans,
            phtncSilePHtrans,
        ]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("CONTCAR", ["", "gz"]))
    def test_vasp_contcar(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileVASP, carSileVASP]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("POSCAR", ["", "gz"]))
    def test_vasp_poscar(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, SileVASP, carSileVASP]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize(
        "sile", _fnames("test", ["xyz", "XYZ", "xyz.gz", "XYZ.gz"])
    )
    def test_xyz(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, xyzSile]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize(
        "sile", _fnames("test", ["molf", "MOLF", "molf.gz", "MOLF.gz"])
    )
    def test_molf(self, sile):
        s = gs(sile)
        for obj in [BaseSile, Sile, moldenSile]:
            assert isinstance(s, obj)

    @pytest.mark.parametrize("sile", _fnames("test", ["win"]))
    def test_wannier90_seed(self, sile):
        sile = gs(sile, cls=SileWannier90)
        for obj in [BaseSile, Sile, SileWannier90, winSileWannier90]:
            assert isinstance(sile, obj)

    @pytest.mark.parametrize("sile", get_siles("write_geometry"))
    def test_write(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        G.set_nsc([1, 1, 1])
        f = sisl_tmp("test_write")

        sile(f, mode="w").write_geometry(G)

    @pytest.mark.parametrize("sile", _my_intersect(["read_lattice"], ["write_lattice"]))
    def test_read_write_lattice(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        L = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc").lattice
        L.set_nsc([1, 1, 1])
        f = sisl_tmp("test_read_write_geom.win")

        # These files does not store the atomic species
        if sys.platform.startswith("win") and issubclass(sile, chgSileVASP):
            pytest.xfail(
                "Windows reading/writing supercell fails for some unknown reason"
            )

        # Write
        sile(f, mode="w").write_lattice(L)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                l = s.read_lattice()
            assert l.equal(L, atol=1e-3)  # pdb files have 8.3 for atomic coordinates
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                l = Lattice.read(s)
            assert l.equal(L, atol=1e-3)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize("sile", _my_intersect(["read_lattice"], ["write_lattice"]))
    def test_read_write_lattice(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        L = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc").lattice
        L.set_nsc([1, 1, 1])
        f = sisl_tmp("test_read_write_geom.win")

        if sys.platform.startswith("win") and issubclass(sile, chgSileVASP):
            pytest.xfail(
                "Windows reading/writing supercell fails for some unknown reason"
            )

        # Write
        sile(f, mode="w").write_lattice(L)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                l = s.read_lattice()
            assert l.equal(L, atol=1e-3)  # pdb files have 8.3 for atomic coordinates
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize(
        "sile", _my_intersect(["read_geometry"], ["write_geometry"])
    )
    def test_read_write_geometry(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        G.set_nsc([1, 1, 1])
        f = sisl_tmp("test_read_write_geom.win")

        if issubclass(sile, (_ncSileTBtrans, deltancSileTBtrans, winSileWannier90)):
            pytest.skip(
                "does not have a proper writing of the atomic species (no direct comparison available)"
            )

        if sys.platform.startswith("win") and issubclass(sile, chgSileVASP):
            pytest.xfail(
                "Windows reading/writing supercell fails for some unknown reason"
            )

        # Write
        sile(f, mode="w").write_geometry(G)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                g = s.read_geometry()
                if isinstance(g, list):
                    g = g[0]
            assert g.equal(
                G, R=False, atol=1e-3
            )  # pdb files have 8.3 for atomic coordinates
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                g = Geometry.read(s)
                if isinstance(g, list):
                    g = g[0]
            assert g.equal(G, R=False, atol=1e-3)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize(
        "sile", _my_intersect(["read_hamiltonian"], ["write_hamiltonian"])
    )
    def test_read_write_hamiltonian(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        if issubclass(sile, _gfSileSiesta):
            pytest.skip("gfSileSiesta handles writing Hamiltonians in a special way")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        H = Hamiltonian(G)
        H.construct([[0.1, 1.45], [0.1, -2.7]])
        f = sisl_tmp("test_read_write_hamiltonian.win")
        # Write
        with sile(f, mode="w") as s:
            s.write_hamiltonian(H)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                h = s.read_hamiltonian()
            assert H.spsame(h)
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                h = Hamiltonian.read(s)
            assert H.spsame(h)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize(
        "sile", _my_intersect(["read_density_matrix"], ["write_density_matrix"])
    )
    def test_read_write_density_matrix(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        DM = DensityMatrix(G, orthogonal=True)
        DM.construct([[0.1, 1.45], [0.1, -2.7]])

        f = sisl_tmp("test_read_write_density_matrix.win")
        # Write
        with sile(f, mode="w") as s:
            s.write_density_matrix(DM)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                dm = s.read_density_matrix(geometry=DM.geometry)
            assert DM.spsame(dm)
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                dm = DensityMatrix.read(s, geometry=DM.geometry)
            assert DM.spsame(dm)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize(
        "sile",
        _my_intersect(["read_energy_density_matrix"], ["write_energy_density_matrix"]),
    )
    def test_read_write_energy_density_matrix(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        EDM = EnergyDensityMatrix(G, orthogonal=True)
        EDM.construct([[0.1, 1.45], [0.1, -2.7]])

        f = sisl_tmp("test_read_write_energy_density_matrix.win")
        # Write
        with sile(f, mode="w") as s:
            s.write_energy_density_matrix(EDM)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                edm = s.read_energy_density_matrix(geometry=EDM.geometry)
            assert EDM.spsame(edm)
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                edm = EnergyDensityMatrix.read(s, geometry=EDM.geometry)
            assert EDM.spsame(edm)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.parametrize(
        "sile", _my_intersect(["read_hamiltonian"], ["write_hamiltonian"])
    )
    def test_read_write_hamiltonian_overlap(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        if issubclass(sile, _gfSileSiesta):
            pytest.skip("gfSileSiesta handles writing Hamiltonians in a special way")

        G = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        H = Hamiltonian(G, orthogonal=False)
        H.construct([[0.1, 1.45], [(0.1, 1), (-2.7, 0.1)]])

        f = sisl_tmp("test_read_write_hamiltonian_overlap.win")
        # Write
        with sile(f, mode="w") as s:
            s.write_hamiltonian(H)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                h = s.read_hamiltonian()
            assert H.spsame(h)
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                h = Hamiltonian.read(s)
            assert H.spsame(h)
        except UnicodeDecodeError as e:
            pass

    @pytest.mark.filterwarnings(
        "ignore", message="*gridncSileSiesta.read_grid cannot determine"
    )
    @pytest.mark.parametrize("sile", _my_intersect(["read_grid"], ["write_grid"]))
    def test_read_write_grid(self, sisl_tmp, sisl_system, sile):
        if issubclass(sile, SileCDF):
            pytest.importorskip("netCDF4")

        g = sisl_system.g.rotate(-30, sisl_system.g.cell[2, :], what="xyz+abc")
        G = Grid([10, 11, 12])
        G[:, :, :] = np.random.rand(10, 11, 12)

        f = sisl_tmp("test_read_write_grid.win")
        # Write
        try:
            with sile(f, mode="w") as s:
                s.write_grid(G)
        except SileError:
            with sile(f, mode="wb") as s:
                s.write_grid(G)
        # Read 1
        try:
            with sile(f, mode="r") as s:
                g = s.read_grid()
            assert np.allclose(g.grid, G.grid, atol=1e-5)
        except UnicodeDecodeError as e:
            pass
        # Read 2
        try:
            with sile(f, mode="r") as s:
                g = Grid.read(s)
            assert np.allclose(g.grid, G.grid, atol=1e-5)
        except UnicodeDecodeError as e:
            pass

    def test_arg_parser1(self, sisl_tmp):
        f = sisl_tmp("something")
        for sile in get_siles(["ArgumentParser"]):
            try:
                sile(f).ArgumentParser()
            except Exception:
                pass

    def test_arg_parser2(self, sisl_tmp):
        f = sisl_tmp("something")
        for sile in get_siles(["ArgumentParser_out"]):
            try:
                sile(f).ArgumentParser()
            except Exception:
                pass


def test_buffer_cls():
    # Check that buffer_cls argument for subclassing classes
    # works correctly
    class customBuffer(BufferSile):
        pass

    class tmpSile(xyzSile, buffer_cls=customBuffer):
        pass

    assert tmpSile._buffer_cls.__name__ == "customBuffer"

    with pytest.raises(TypeError):

        class tmpSile(xyzSile, buffer_cls=object):
            pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.mixing import AdaptiveDIISMixer, DIISMixer

pytestmark = pytest.mark.mixing


@pytest.mark.parametrize("history", [2, 10])
def test_diis_mixer(history):
    # test for different history lengths
    def scf(f):
        return np.cos(f)

    f = np.linspace(0, 7, 1000)
    mix = DIISMixer(history=history)
    s = str(mix)

    dmax = 1
    i = 0
    while dmax > 1e-6:
        i += 1
        df = scf(f) - f
        dmax = np.fabs(df).max()
        f = mix(f, df)


@pytest.mark.parametrize("history", [2, 10])
@pytest.mark.parametrize("weight", [0.5, (0.3, 0.7)])
def test_adiis_mixer(weight, history):
    # test for different history lengths
    def scf(f):
        return np.cos(f)

    f = np.linspace(0, 7, 1000)
    mix = AdaptiveDIISMixer(weight, history=history)

    dmax = 1
    i = 0
    while dmax > 1e-6:
        i += 1
        df = scf(f) - f
        dmax = np.fabs(df).max()
        f = mix(f, df)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.tbtrans]

netCDF4 = pytest.importorskip("netCDF4")


@pytest.mark.slow
def test_2_projection_content(sisl_files):
    tbt = sisl.get_sile(
        sisl_files("siesta", "tbtrans", "c60_projection", "projection.TBT.nc")
    )
    tbtp = sisl.get_sile(
        sisl_files("siesta", "tbtrans", "c60_projection", "projection.TBT.Proj.nc")
    )

    assert np.allclose(tbt.E, tbtp.E)
    assert np.allclose(tbt.kpt, tbtp.kpt)

    # Get geometry
    assert tbt.read_geometry() == tbtp.read_geometry()

    assert np.allclose(tbt.n_btd(), tbtp.n_btd())
    assert np.allclose(tbt.pivot(), tbtp.pivot())

    # Loop molecules
    left = tbt.elecs[0]
    right = tbt.elecs[1]
    for mol in tbtp.molecules:
        for proj in tbtp.projections(mol):
            t1 = tbtp.transmission((left, mol, proj), (right, mol, proj))
            t2 = tbtp.transmission(
                ".".join((left, mol, proj)), ".".join((right, mol, proj))
            )
            assert np.allclose(t1, t2)

            te1 = tbtp.transmission_eig((left, mol, proj), (right, mol, proj))
            te2 = tbtp.transmission_eig(
                ".".join((left, mol, proj)), ".".join((right, mol, proj))
            )
            assert np.allclose(te1, te2)
            assert np.allclose(t1, te1.sum(-1))
            assert np.allclose(t2, te2.sum(-1))

    # Check eigenstate
    es = tbtp.eigenstate("C60")
    assert len(es) == 3  # 1-HOMO, 2-LUMO
    assert (es.eig < 0.0).nonzero()[0].size == 1
    assert (es.eig > 0.0).nonzero()[0].size == 2
    assert np.allclose(es.norm2(), 1)


@pytest.mark.slow
def test_2_projection_tbtav(sisl_files, sisl_tmp):
    tbtp = sisl.get_sile(
        sisl_files("siesta", "tbtrans", "c60_projection", "projection.TBT.Proj.nc")
    )
    f = sisl_tmp("2_projection.TBT.Proj.AV.nc")
    tbtp.write_tbtav(f)


def test_2_projection_ArgumentParser(sisl_files, sisl_tmp):
    pytest.importorskip("matplotlib", reason="matplotlib not available")
    import argparse

    # Local routine to run the collected actions
    def run(ns):
        ns._actions_run = True
        # Run all so-far collected actions
        for A, Aargs, Akwargs in ns._actions:
            A(*Aargs, **Akwargs)
        ns._actions_run = False
        ns._actions = []

    tbtp = sisl.get_sile(
        sisl_files("siesta", "tbtrans", "c60_projection", "projection.TBT.Proj.nc")
    )

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    p.parse_args([], namespace=ns)

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--energy", " -1.995:1.995"], namespace=ns)
    assert not out._actions_run
    run(out)

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--norm", "orbital"], namespace=ns)
    run(out)
    assert out._norm == "orbital"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--norm", "atom"], namespace=ns)
    run(out)
    assert out._norm == "atom"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--atom", "10:11,14"], namespace=ns)
    run(out)
    assert out._Ovalue == "10:11,14"
    # Only atom 14 is in the device region
    assert np.all(out._Orng + 1 == [14])

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--atom", "10:11,12,14:20"], namespace=ns)
    run(out)
    assert out._Ovalue == "10:11,12,14:20"
    # Only 13-72 is in the device
    assert np.all(out._Orng + 1 == [14, 15, 16, 17, 18, 19, 20])

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        ["--transmission", "Left.C60.HOMO", "Right.C60.HOMO"], namespace=ns
    )
    run(out)
    assert len(out._data) == 2
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][0] == "T"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--molecules", "-P", "C60"], namespace=ns)
    run(out)

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        [
            "--transmission",
            "Left",
            "Right.C60.LUMO",
            "--transmission",
            "Left.C60.LUMO",
            "Right",
        ],
        namespace=ns,
    )
    run(out)
    assert len(out._data) == 3
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][0] == "T"
    assert out._data_header[2][0] == "T"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        ["--ados", "Left.C60.HOMO", "--ados", "Left.C60.LUMO"], namespace=ns
    )
    run(out)
    assert len(out._data) == 3
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][:2] == "AD"
    assert out._data_header[2][:2] == "AD"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        ["--transmission-eig", "Left.C60.HOMO", "Right.C60.LUMO"], namespace=ns
    )
    run(out)
    assert out._data_header[0][0] == "E"
    for i in range(1, len(out._data)):
        assert out._data_header[i][:4] == "Teig"

    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(["--info"], namespace=ns)

    # Test output
    f = sisl_tmp("projection.dat")
    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        ["--transmission-eig", "Left", "Right.C60.HOMO", "--out", f], namespace=ns
    )
    assert len(out._data) == 0

    f1 = sisl_tmp("projection_1.dat")
    f2 = sisl_tmp("projection_2.dat")
    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        [
            "--transmission",
            "Left",
            "Right.C60.HOMO",
            "--out",
            f1,
            "--ados",
            "Left.C60.HOMO",
            "--atom",
            "13:2:72",
            "--ados",
            "Left.C60.HOMO",
            "--atom",
            "14:2:72",
            "--ados",
            "Left.C60.HOMO",
            "--out",
            f2,
        ],
        namespace=ns,
    )

    d = sisl.io.tableSile(f1).read_data()
    assert len(d) == 2
    d = sisl.io.tableSile(f2).read_data()
    assert len(d) == 4
    assert np.allclose(d[1, :], d[2, :] + d[3, :])

    f = sisl_tmp("projection_T.png")
    p, ns = tbtp.ArgumentParser(argparse.ArgumentParser(conflict_handler="resolve"))
    out = p.parse_args(
        [
            "--transmission",
            "Left",
            "Right.C60.HOMO",
            "--transmission",
            "Left.C60.HOMO",
            "Right.C60.HOMO",
            "--plot",
            f,
        ],
        namespace=ns,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from functools import partial

import pytest

from sisl.utils.ranges import *

pytestmark = pytest.mark.utils


@pytest.mark.ranges
class TestRanges:
    def test_strseq_colon(self):
        ranges = strseq(int, "1:2:5")
        assert ranges == (1, 2, 5)
        ranges = strseq(int, "1:2")
        assert ranges == (1, 2)
        ranges = strseq(int, "1:")
        assert ranges == (1, None)
        ranges = strseq(int, ":4")
        assert ranges == (None, 4)
        ranges = strseq(int, ":4:")
        assert ranges == (None, 4, None)

    def test_strseq_minus(self):
        ranges = strseq(int, "1-2-5")
        assert ranges == (1, 2, 5)
        ranges = strseq(int, "1-2")
        assert ranges == (1, 2)
        ranges = strseq(int, "1-")
        assert ranges == (1, None)
        ranges = strseq(int, "-4")
        assert ranges == -4
        ranges = strseq(int, "-4", end=5)
        assert ranges == 1
        ranges = strseq(int, "-4-")
        assert ranges == (-4, None)

    def test_strmap1(self):
        assert strmap(int, "1") == [1]
        assert strmap(int, "") == [None]
        assert strmap(int, "1-") == [(1, None)]
        assert strmap(int, "-") == [(None, None)]
        assert strmap(int, "-1") == [-1]
        assert strmap(int, "-1", start=1) == [-1]
        assert strmap(int, "-", start=1, end=2) == [(1, 2)]
        assert strmap(int, "-1:2") == [(-1, 2)]
        assert strmap(int, "1,2") == [1, 2]
        assert strmap(int, "1,2[0,2-]") == [1, (2, [0, (2, None)])]
        assert strmap(int, "1,2-[0,-2]") == [1, ((2, None), [0, -2])]
        assert strmap(int, "1,2[0,2]") == [1, (2, [0, 2])]
        assert strmap(int, "1,2-3[0,2]") == [1, ((2, 3), [0, 2])]
        assert strmap(int, "1,[2,3][0,2]") == [1, (2, [0, 2]), (3, [0, 2])]
        assert strmap(int, "[82][10]") == [(82, [10])]
        assert strmap(int, "[82,83][10]") == [(82, [10]), (83, [10])]
        assert strmap(int, "[82,83][10-13]") == [(82, [(10, 13)]), (83, [(10, 13)])]

    def test_strmap2(self):
        with pytest.raises(ValueError):
            strmap(int, "1", sep="*")

    def test_strmap3(self):
        sm = partial(strmap, sep="c")
        assert sm(int, "1") == [1]
        assert sm(int, "1,2") == [1, 2]
        assert sm(int, "1,2{0,2}") == [1, (2, [0, 2])]
        assert sm(int, "1,2-3{0,2}") == [1, ((2, 3), [0, 2])]
        assert sm(int, "1,{2,3}{0,2}") == [1, (2, [0, 2]), (3, [0, 2])]
        assert sm(int, "{82}{10}") == [(82, [10])]
        assert sm(int, "{82,83}{10}") == [(82, [10]), (83, [10])]
        assert sm(int, "{82,83}{10-13}") == [(82, [(10, 13)]), (83, [(10, 13)])]

    def test_strmap4(self):
        with pytest.raises(ValueError):
            strmap(int, "1[oestuh]]")

    def test_strmap5(self):
        r = strmap(int, "1-", end=5)
        assert r == [(1, 5)]
        r = strmap(int, "1-", start=0, end=5)
        assert r == [(1, 5)]
        r = strmap(int, "-4", start=0, end=5)
        assert r == [1]
        r = strmap(int, "-", start=0, end=5)
        assert r == [(0, 5)]

    def test_lstranges1(self):
        ranges = strmap(int, "1,2-3[0,2]")
        assert lstranges(ranges) == [1, [2, [0, 2]], [3, [0, 2]]]
        ranges = strmap(int, "1,2-4[0-4,2],6[1-3],9-10")
        assert lstranges(ranges) == [
            1,
            [2, [0, 1, 2, 3, 4, 2]],
            [3, [0, 1, 2, 3, 4, 2]],
            [4, [0, 1, 2, 3, 4, 2]],
            [6, [1, 2, 3]],
            9,
            10,
        ]
        ranges = strmap(int, "1,[2,4][0-4,2],6[1-3],9-10")
        assert lstranges(ranges) == [
            1,
            [2, [0, 1, 2, 3, 4, 2]],
            [4, [0, 1, 2, 3, 4, 2]],
            [6, [1, 2, 3]],
            9,
            10,
        ]
        ranges = strmap(int, "1,[2,4,6-7][0,3-4,2],6[1-3],9-10")
        assert lstranges(ranges) == [
            1,
            [2, [0, 3, 4, 2]],
            [4, [0, 3, 4, 2]],
            [6, [0, 3, 4, 2]],
            [7, [0, 3, 4, 2]],
            [6, [1, 2, 3]],
            9,
            10,
        ]
        ranges = strmap(int, "[82,83][10-13]")
        assert lstranges(ranges) == [[82, [10, 11, 12, 13]], [83, [10, 11, 12, 13]]]
        ranges = strmap(int, " [82,85][3]")
        assert lstranges(ranges) == [[82, [3]], [85, [3]]]
        ranges = strmap(int, "81,[82,85][3]")
        assert lstranges(ranges) == [81, [82, [3]], [85, [3]]]

    def test_lstranges2(self):
        ranges = strmap(int, "1:2:5")
        assert lstranges(ranges) == [1, 3, 5]
        ranges = strmap(int, "1-2-5")
        assert lstranges(ranges) == [1, 3, 5]

    def test_fileindex1(self):
        fname = "hello[1]"
        assert fileindex("hehlo")[1] is None
        assert fileindex(fname)[1] == 1
        assert fileindex("hehlo[1,2]")[1] == [1, 2]
        assert fileindex("hehlo[1-2]")[1] == [1, 2]
        assert fileindex("hehlo[1[1],2]")[1] == [[1, [1]], 2]

    def test_list2str(self):
        a = list2str([2, 4, 5, 6])
        assert a == "2, 4-6"
        a = list2str([2, 4, 5, 6, 8, 9])
        assert a == "2, 4-6, 8-9"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.physics.distribution import *

pytestmark = [pytest.mark.physics, pytest.mark.distribution]


def test_distribution1():
    x = np.linspace(-2, 2, 10000)
    dx = x[1] - x[0]
    d = get_distribution("gaussian", smearing=0.025)
    assert d(x).sum() * dx == pytest.approx(1, abs=1e-6)
    d = get_distribution("lorentzian", smearing=1e-3)
    assert d(x).sum() * dx == pytest.approx(1, abs=1e-3)


def test_distribution2():
    x = np.linspace(-2, 2, 10000)
    d = get_distribution("gaussian", smearing=0.025)
    assert np.allclose(d(x), gaussian(x, 0.025))
    d = get_distribution("lorentzian", smearing=1e-3)
    assert np.allclose(d(x), lorentzian(x, 1e-3))
    d = get_distribution("step", smearing=1e-3)
    assert np.allclose(d(x), step_function(x))
    d = get_distribution("heaviside", smearing=1e-3, x0=1)
    assert np.allclose(d(x), 1 - step_function(x, 1))
    d = get_distribution("heaviside", x0=-0.5)
    assert np.allclose(d(x), heaviside(x, -0.5))


def test_distribution3():
    with pytest.raises(ValueError):
        get_distribution("unknown-function")


def test_distribution_x0():
    x1 = np.linspace(-2, 2, 10000)
    x2 = np.linspace(-3, 1, 10000)
    d1 = get_distribution("gaussian")
    d2 = get_distribution("gaussian", x0=-1)
    assert np.allclose(d1(x1), d2(x2))
    d1 = get_distribution("lorentzian")
    d2 = get_distribution("lorentzian", x0=-1)
    assert np.allclose(d1(x1), d2(x2))


def test_fermi_dirac():
    E1 = np.linspace(-2, 2, 1000)
    E2 = np.linspace(-3, 1, 1000)
    assert np.allclose(fermi_dirac(E1, 0.1), fermi_dirac(E2, 0.1, -1))


def test_bose_einstein():
    E1 = np.linspace(-2, 2, 1000)
    E2 = np.linspace(-3, 1, 1000)
    assert np.allclose(bose_einstein(E1, 0.1), bose_einstein(E2, 0.1, -1))


def test_cold():
    E1 = np.linspace(-2, 2, 1000)
    E2 = np.linspace(-3, 1, 1000)
    assert np.allclose(cold(E1, 0.1), cold(E2, 0.1, -1))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import (
    Atom,
    AtomicOrbital,
    DensityMatrix,
    Geometry,
    Grid,
    Lattice,
    SparseAtom,
    SparseOrbital,
    SphericalOrbital,
    Spin,
)


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.bond = bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )

            n = 60
            rf = np.linspace(0, bond * 1.01, n)
            rf = (rf, rf)
            orb = SphericalOrbital(1, rf, 2.0)
            C = Atom(6, orb.toAtomicOrbital())
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )
            self.D = DensityMatrix(self.g)
            self.DS = DensityMatrix(self.g, orthogonal=False)

            def func(D, ia, atoms, atoms_xyz):
                idx = D.geometry.close(
                    ia, R=(0.1, 1.44), atoms=atoms, atoms_xyz=atoms_xyz
                )
                ia = ia * 3

                i0 = idx[0] * 3
                i1 = idx[1] * 3
                # on-site
                p = 1.0
                D.D[ia, i0] = p
                D.D[ia + 1, i0 + 1] = p
                D.D[ia + 2, i0 + 2] = p

                # nn
                p = 0.1

                # on-site directions
                D.D[ia, ia + 1] = p
                D.D[ia, ia + 2] = p
                D.D[ia + 1, ia] = p
                D.D[ia + 1, ia + 2] = p
                D.D[ia + 2, ia] = p
                D.D[ia + 2, ia + 1] = p

                D.D[ia, i1 + 1] = p
                D.D[ia, i1 + 2] = p

                D.D[ia + 1, i1] = p
                D.D[ia + 1, i1 + 2] = p

                D.D[ia + 2, i1] = p
                D.D[ia + 2, i1 + 1] = p

            self.func = func

    return t()


@pytest.fixture(
    scope="module",
    params=["direct", "pre-compute"],
)
def density_method(request):
    return request.param


@pytest.mark.physics
@pytest.mark.densitymatrix
class TestDensityMatrix:
    def test_objects(self, setup):
        assert len(setup.D.xyz) == 2
        assert setup.g.no == len(setup.D)

    def test_dtype(self, setup):
        assert setup.D.dtype == np.float64

    def test_ortho(self, setup):
        assert setup.D.orthogonal

    def test_set1(self, setup):
        D = setup.D.copy()
        D.D[0, 0] = 1.0
        assert D[0, 0] == 1.0
        assert D[1, 0] == 0.0

    def test_mulliken(self, setup):
        D = setup.D.copy()
        D.construct(setup.func)
        mulliken = D.mulliken("atom")
        assert mulliken.shape == (len(D.geometry),)
        mulliken = D.mulliken("orbital")
        assert mulliken.shape == (len(D),)

    def test_mulliken_values_orthogonal(self, setup):
        D = setup.D.copy()
        D.D[0, 0] = 1.0
        D.D[1, 1] = 2.0
        D.D[1, 2] = 2.0
        mulliken = D.mulliken("orbital")
        assert np.allclose(mulliken[:2], [1.0, 2.0])
        assert mulliken.sum() == pytest.approx(3)
        mulliken = D.mulliken("atom")
        assert mulliken[0] == pytest.approx(3)
        assert mulliken.sum() == pytest.approx(3)

    def test_mulliken_values_non_orthogonal(self, setup):
        D = setup.DS.copy()
        D[0, 0] = (1.0, 1.0)
        D[1, 1] = (2.0, 1.0)
        D[1, 2] = (2.0, 0.5)
        mulliken = D.mulliken("orbital")
        assert np.allclose(mulliken[:2], [1.0, 3.0])
        assert mulliken.sum() == pytest.approx(4)
        mulliken = D.mulliken("atom")
        assert mulliken[0] == pytest.approx(4)
        assert mulliken.sum() == pytest.approx(4)

    def test_mulliken_polarized(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("P"))
        # 1 charge onsite for each spin-up
        # 0.5 charge onsite for each spin-down
        D.construct([[0.1, bond + 0.01], [(1.0, 0.5), (0.1, 0.1)]])

        m = D.mulliken("orbital")
        assert m[0].sum() == pytest.approx(3)
        assert m[1].sum() == pytest.approx(1)
        m = D.mulliken("atom")
        assert m[0].sum() == pytest.approx(3)
        assert m[1].sum() == pytest.approx(1)

    @pytest.mark.parametrize("method", ["wiberg", "mayer"])
    @pytest.mark.parametrize("option", ["", ":spin"])
    @pytest.mark.parametrize("projection", ["atom", "orbitals"])
    def test_bond_order(self, setup, method, option, projection):
        D = setup.D.copy()
        D.construct(setup.func)
        BO = D.bond_order(method + option, projection)
        if projection == "atom":
            assert isinstance(BO, SparseAtom)
            assert BO.shape[:2] == (D.geometry.na, D.geometry.na_s)
        elif projection == "orbitals":
            assert isinstance(BO, SparseOrbital)
            assert BO.shape[:2] == (D.geometry.no, D.geometry.no_s)

    def test_rho1(self, setup, density_method):
        D = setup.D.copy()
        D.construct(setup.func)
        grid = Grid(0.2, geometry=setup.D.geometry)
        D.density(grid, method=density_method)

    @pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
    def test_rho2(self, density_method):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        n = 60
        rf = np.linspace(0, bond * 1.01, n)
        rf = (rf, rf)
        orb = SphericalOrbital(1, rf, 2.0)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g)
        D.construct([[0.1, bond + 0.01], [1.0, 0.1]])
        grid = Grid(0.2, geometry=D.geometry)
        D.density(grid, method=density_method)

        D = DensityMatrix(g, spin=Spin("P"))
        D.construct([[0.1, bond + 0.01], [(1.0, 0.5), (0.1, 0.1)]])
        grid = Grid(0.2, geometry=D.geometry)
        D.density(grid, method=density_method)
        D.density(grid, [1.0, -1], method=density_method)
        D.density(grid, 0, method=density_method)
        D.density(grid, 1, method=density_method)

        D = DensityMatrix(g, spin=Spin("NC"))
        D.construct(
            [[0.1, bond + 0.01], [(1.0, 0.5, 0.01, 0.01), (0.1, 0.1, 0.1, 0.1)]]
        )
        grid = Grid(0.2, geometry=D.geometry)
        D.density(grid, method=density_method)
        D.density(grid, [[1.0, 0.0], [0.0, -1]], method=density_method)

        D = DensityMatrix(g, spin=Spin("SO"))
        D.construct(
            [
                [0.1, bond + 0.01],
                [
                    (1.0, 0.5, 0.01, 0.01, 0.01, 0.01, 0.0, 0.0),
                    (0.1, 0.1, 0.1, 0.1, 0.0, 0.0, 0.0, 0.0),
                ],
            ]
        )
        grid = Grid(0.2, geometry=D.geometry)
        D.density(grid, method=density_method)
        D.density(grid, [[1.0, 0.0], [0.0, -1]], method=density_method)
        D.density(grid, Spin.X, method=density_method)
        D.density(grid, Spin.Y, method=density_method)
        D.density(grid, Spin.Z, method=density_method)

    @pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
    def test_orbital_momentum(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("SO"))
        D.construct(
            [
                [0.1, bond + 0.01],
                [
                    (1.0, 0.5, 0.01, 0.01, 0.01, 0.01, 0.0, 0.0),
                    (0.1, 0.1, 0.1, 0.1, 0.0, 0.0, 0.0, 0.0),
                ],
            ]
        )
        D.orbital_momentum("atom")
        D.orbital_momentum("orbital")

    def test_spin_align_pol(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("p"))
        D.construct([[0.1, bond + 0.01], [(1.0, 0.5), (0.1, 0.2)]])
        D_mull = D.mulliken()
        assert D_mull.shape == (2, len(D))

        v = np.array([1, 2, 3])
        d = D.spin_align(v)
        d_mull = d.mulliken()
        assert np.allclose(d_mull, d.astype(np.complex128).mulliken())
        assert d_mull.shape == (4, len(D))

        assert not np.allclose(D_mull[1], d_mull[3])
        assert np.allclose(D_mull[0], d_mull[0])

    def test_spin_align_nc(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("nc"))
        D.construct(
            [[0.1, bond + 0.01], [(1.0, 0.5, 0.01, 0.01), (0.1, 0.2, 0.1, 0.1)]]
        )
        D_mull = D.mulliken()
        v = np.array([1, 2, 3])
        d = D.spin_align(v)
        d_mull = d.mulliken()
        assert np.allclose(d_mull, d.astype(np.complex128).mulliken())
        assert not np.allclose(D_mull, d_mull)
        assert np.allclose(D_mull[0], d_mull[0])

    @pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
    def test_spin_align_so(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("SO"))
        D.construct(
            [
                [0.1, bond + 0.01],
                [
                    (1.0, 0.5, 0.01, 0.01, 0.01, 0.01, 0.2, 0.2),
                    (0.1, 0.2, 0.1, 0.1, 0.0, 0.1, 0.2, 0.3),
                ],
            ]
        )
        D_mull = D.mulliken()
        v = np.array([1, 2, 3])
        d = D.spin_align(v, atoms=0)
        d_mull = d.mulliken()
        assert np.allclose(d_mull, d.astype(np.complex128).mulliken())
        assert not np.allclose(D_mull, d_mull)
        assert np.allclose(D_mull[0], d_mull[0])

    def test_spin_rotate_pol(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("p"))
        D.construct([[0.1, bond + 0.01], [(1.0, 0.5), (0.1, 0.2)]])

        D_mull = D.mulliken()
        assert np.allclose(D_mull, D.astype(np.complex128).mulliken())
        assert D_mull.shape == (2, len(D))

        d = D.spin_rotate([45, 60, 90], rad=False)
        d_mull = d.mulliken()
        assert d_mull.shape == (4, len(D))

        assert not np.allclose(D_mull[1], d_mull[3])
        assert np.allclose(D_mull[0], d_mull[0])

    def test_spin_rotate_pol_full(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("p"))
        D.construct([[0.1, bond + 0.01], [(1.0, 0), (0.1, 0.0)]])

        D_mull = D.mulliken()
        assert np.allclose(D_mull, D.astype(np.complex128).mulliken())
        assert D_mull.shape == (2, len(D))

        # Euler (noop)
        d = D.spin_rotate([0, 0, 64], rad=False)
        assert d.spin.is_polarized
        assert np.allclose(d.mulliken()[1], D.mulliken()[1])
        d = D.spin_rotate([180, 180, 64], rad=False)
        assert d.spin.is_polarized
        assert np.allclose(d.mulliken()[1], D.mulliken()[1])

        # Euler (full)
        d = D.spin_rotate([0, 180, 64], rad=False)
        assert d.spin.is_polarized
        assert np.allclose(d.mulliken()[1], -D.mulliken()[1])
        d = D.spin_rotate([180, 0, 64], rad=False)
        assert d.spin.is_polarized
        assert np.allclose(d.mulliken()[1], -D.mulliken()[1])

    def test_spin_rotate_nc(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("nc"))
        D.construct(
            [[0.1, bond + 0.01], [(1.0, 0.5, 0.01, 0.01), (0.1, 0.2, 0.1, 0.1)]]
        )

        D_mull = D.mulliken()
        assert np.allclose(D_mull, D.astype(np.complex128).mulliken())
        d = D.spin_rotate([45, 60, 90], rad=False)

        d_mull = d.mulliken()

        assert not np.allclose(D_mull, d_mull)
        assert np.allclose(D_mull[0], d_mull[0])

    @pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
    def test_spin_rotate_so(self):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        orb = AtomicOrbital("px", R=bond * 1.001)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )
        D = DensityMatrix(g, spin=Spin("SO"))
        D.construct(
            [
                [0.1, bond + 0.01],
                [
                    (1.0, 0.5, 0.01, 0.01, 0.01, 0.01, 0.2, 0.2),
                    (0.1, 0.2, 0.1, 0.1, 0.0, 0.1, 0.2, 0.3),
                ],
            ]
        )
        D_mull = D.mulliken()
        assert np.allclose(D_mull, D.astype(np.complex128).mulliken())
        d = D.spin_rotate([45, 60, 90], rad=False)
        d_mull = d.mulliken()
        assert not np.allclose(D_mull, d_mull)
        assert np.allclose(D_mull[0], d_mull[0])

        d1 = (
            d.spin_rotate([0, 0, -90], rad=False)
            .spin_rotate([0, -60, 0], rad=False)
            .spin_rotate([-45, 0, 0], rad=False)
        )
        d1_mull = d1.mulliken()
        assert not np.allclose(d_mull, d1_mull)
        assert np.allclose(D_mull, d1_mull)

    def test_rho_eta(self, setup, density_method):
        D = setup.D.copy()
        D.construct(setup.func)
        grid = Grid(0.2, geometry=setup.D.geometry)
        D.density(grid, eta=True, method=density_method)

    def test_rho_smaller_grid1(self, setup, density_method):
        D = setup.D.copy()
        D.construct(setup.func)
        lattice = setup.D.geometry.cell.copy() / 2
        grid = Grid(0.2, geometry=setup.D.geometry.copy(), lattice=lattice)
        D.density(grid, method=density_method)

    def test_rho_fail_p(self, density_method):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        n = 60
        rf = np.linspace(0, bond * 1.01, n)
        rf = (rf, rf)
        orb = SphericalOrbital(1, rf, 2.0)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )

        D = DensityMatrix(g, spin=Spin("P"))
        D.construct([[0.1, bond + 0.01], [(1.0, 0.5), (0.1, 0.1)]])
        grid = Grid(0.2, geometry=D.geometry)
        with pytest.raises(ValueError):
            D.density(grid, [1.0, -1, 0.0], method=density_method)

    def test_rho_fail_nc(self, density_method):
        bond = 1.42
        sq3h = 3.0**0.5 * 0.5
        lattice = Lattice(
            np.array(
                [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
            )
            * bond,
            nsc=[3, 3, 1],
        )

        n = 60
        rf = np.linspace(0, bond * 1.01, n)
        rf = (rf, rf)
        orb = SphericalOrbital(1, rf, 2.0)
        C = Atom(6, orb)
        g = Geometry(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
            atoms=C,
            lattice=lattice,
        )

        D = DensityMatrix(g, spin=Spin("NC"))
        D.construct(
            [[0.1, bond + 0.01], [(1.0, 0.5, 0.01, 0.01), (0.1, 0.1, 0.1, 0.1)]]
        )
        grid = Grid(0.2, geometry=D.geometry)
        with pytest.raises(ValueError):
            D.density(grid, [1.0, 0.0], method=density_method)

    def test_pickle(self, setup):
        import pickle as p

        D = setup.D.copy()
        D.construct(setup.func)
        s = p.dumps(D)
        d = p.loads(s)
        assert d.spsame(D)
        assert np.allclose(d.eigh(), D.eigh())

    def test_transform(self, setup):
        D = DensityMatrix(setup.g, spin="so")
        a = np.arange(8)
        for ia in setup.g:
            D[ia, ia] = a
        Dcsr = [D.tocsr(i) for i in range(D.shape[2])]

        Dt = D.transform(spin="unpolarized").astype(np.float32)
        assert np.abs(0.5 * Dcsr[0] + 0.5 * Dcsr[1] - Dt.tocsr(0)).sum() == 0

        Dt = D.transform(spin="polarized", orthogonal=False)
        assert np.abs(Dcsr[0] - Dt.tocsr(0)).sum() == 0
        assert np.abs(Dcsr[1] - Dt.tocsr(1)).sum() == 0
        assert np.abs(Dt.tocsr(2)).sum() != 0

        Dt = D.transform(spin="non-colinear", orthogonal=False)
        assert np.abs(Dcsr[0] - Dt.tocsr(0)).sum() == 0
        assert np.abs(Dcsr[1] - Dt.tocsr(1)).sum() == 0
        assert np.abs(Dcsr[2] - Dt.tocsr(2)).sum() == 0
        assert np.abs(Dcsr[3] - Dt.tocsr(3)).sum() == 0
        assert np.abs(Dt.tocsr(-1)).sum() != 0

    def test_transform_nonortho(self, setup):
        D = DensityMatrix(setup.g, spin="polarized", orthogonal=False)
        a = np.arange(3)
        a[-1] = 1.0
        for ia in setup.g:
            D[ia, ia] = a

        Dt = D.transform(spin="unpolarized").astype(np.float32)
        assert np.abs(0.5 * D.tocsr(0) + 0.5 * D.tocsr(1) - Dt.tocsr(0)).sum() == 0
        assert np.abs(D.tocsr(-1) - Dt.tocsr(-1)).sum() == 0
        Dt = D.transform(spin="polarized")
        assert np.abs(D.tocsr(0) - Dt.tocsr(0)).sum() == 0
        assert np.abs(D.tocsr(1) - Dt.tocsr(1)).sum() == 0
        Dt = D.transform(spin="polarized", orthogonal=True)
        assert np.abs(D.tocsr(0) - Dt.tocsr(0)).sum() == 0
        assert np.abs(D.tocsr(1) - Dt.tocsr(1)).sum() == 0
        Dt = D.transform(spin="non-colinear", orthogonal=False)
        assert np.abs(D.tocsr(0) - Dt.tocsr(0)).sum() == 0
        assert np.abs(D.tocsr(1) - Dt.tocsr(1)).sum() == 0
        assert np.abs(Dt.tocsr(2)).sum() == 0
        assert np.abs(Dt.tocsr(3)).sum() == 0
        assert np.abs(D.tocsr(-1) - Dt.tocsr(-1)).sum() == 0
        Dt = D.transform(spin="so", orthogonal=True)
        assert np.abs(D.tocsr(0) - Dt.tocsr(0)).sum() == 0
        assert np.abs(D.tocsr(1) - Dt.tocsr(1)).sum() == 0
        assert np.abs(Dt.tocsr(-1)).sum() == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.siesta.fa import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_pdos_kgrid_fa(sisl_files):
    f = sisl_files("siesta", "Si_pdos_k", "Si_pdos.FA")
    fa = faSileSiesta(f).read_data()

    assert len(fa) == 2
    fa1 = faSileSiesta(f).read_force()
    assert np.allclose(fa, fa1)


def test_read_write_fa(sisl_tmp):
    f = sisl_tmp("test.FA")

    fa = np.random.rand(10, 3)
    faSileSiesta(f, "w").write_force(fa)
    fa2 = faSileSiesta(f).read_force()

    assert len(fa) == len(fa2)
    assert np.allclose(fa, fa2)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.viz.processors.math import normalize

pytestmark = [pytest.mark.viz, pytest.mark.processors]


def test_normalize():
    data = [0, 1, 2]

    assert np.allclose(normalize(data), [0, 0.5, 1])

    assert np.allclose(normalize(data, vmin=-1, vmax=1), [-1, 0, 1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
from scipy.sparse import csr_matrix

from sisl import (
    Atom,
    Cuboid,
    Ellipsoid,
    Geometry,
    Grid,
    Lattice,
    SislError,
    SislWarning,
    geom,
)

pytestmark = [pytest.mark.grid]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            alat = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * alat,
                nsc=[3, 3, 1],
            )
            self.g = Grid([10, 10, 100], lattice=self.lattice)
            self.g.fill(2.0)

    return t()


@pytest.mark.grid
class TestGrid:
    def test_print(self, setup):
        str(setup.g)

    def test_init(self, setup):
        Grid(0.1, lattice=setup.lattice)

    def test_append(self, setup):
        g = setup.g.append(setup.g, 0)
        assert np.allclose(g.grid.shape, [20, 10, 100])
        g = setup.g.append(setup.g, 1)
        assert np.allclose(g.grid.shape, [10, 20, 100])
        g = setup.g.append(setup.g, 2)
        assert np.allclose(g.grid.shape, [10, 10, 200])

    def test_set(self, setup):
        v = setup.g[0, 0, 0]
        setup.g[0, 0, 0] = 3
        assert setup.g.grid[0, 0, 0] == 3
        assert setup.g[0, 0, 0] == 3
        setup.g[0, 0, 0] = v
        assert setup.g[0, 0, 0] == v

    def test_size(self, setup):
        assert np.allclose(setup.g.grid.shape, [10, 10, 100])

    def test_item(self, setup):
        assert np.allclose(setup.g[1:2, 1:2, 2:3], setup.g.grid[1:2, 1:2, 2:3])

    def test_dcell(self, setup):
        assert np.all(setup.g.dcell * setup.g.cell >= 0)

    def test_dvolume(self, setup):
        assert setup.g.dvolume > 0

    def test_shape(self, setup):
        assert np.all(setup.g.shape == setup.g.grid.shape)

    def test_dtype(self, setup):
        assert setup.g.dtype == setup.g.grid.dtype

    def test_copy(self, setup):
        assert setup.g.copy() == setup.g
        assert not setup.g.copy() != setup.g
        gc = setup.g.copy(dtype=np.complex128)
        assert gc.dtype == np.complex128

    def test_add1(self, setup):
        g = setup.g + setup.g
        assert np.allclose(g.grid, (setup.g * 2).grid)
        g = setup.g.copy()
        g *= 2
        assert np.allclose(g.grid, (setup.g * 2).grid)
        g = setup.g.copy()
        g /= 2
        assert np.allclose(g.grid, (setup.g / 2).grid)

    def test_add2(self, setup):
        g = setup.g + 2.0
        assert np.allclose(g.grid, setup.g.grid + 2)
        g = setup.g.copy()
        g += 2.0
        g -= 2.0
        assert np.allclose(g.grid, setup.g.grid)
        g = setup.g + setup.g
        assert np.allclose(g.grid, setup.g.grid * 2)
        assert np.allclose((g - setup.g).grid, setup.g.grid)

    def test_add_fail1(self, setup):
        g = Grid(np.array(setup.g.shape) // 2 + 1, lattice=setup.g.lattice.copy())
        with pytest.raises(ValueError):
            setup.g + g

    def test_iadd1(self):
        g = Grid([10, 10, 10])
        g.fill(1)
        old = g.copy()
        g += g
        g -= old
        assert np.allclose(g.grid, 1)
        g -= g
        assert np.allclose(g.grid, 0)

    def test_op1(self, setup):
        g = setup.g * setup.g
        assert np.allclose(g.grid, setup.g.grid * setup.g.grid)
        g = setup.g.copy()
        g *= setup.g
        assert np.allclose(g.grid, setup.g.grid * setup.g.grid)
        g = setup.g * setup.g
        g /= setup.g
        assert np.allclose(g.grid, setup.g.grid)

    def test_swapaxes(self, setup):
        g = setup.g.swapaxes(0, 1)
        assert np.allclose(setup.g.cell[0, :], g.cell[1, :])
        assert np.allclose(setup.g.cell[1, :], g.cell[0, :])

    def test_average(self, setup):
        g = setup.g.copy()
        g.grid.fill(1)
        shape = g.shape
        for i in range(3):
            assert g.average(i).grid.sum() == shape[0] * shape[1] * shape[2] / shape[i]
        assert g.average(0).shape == (1, shape[1], shape[2])
        assert g.average(1).shape == (shape[0], 1, shape[2])
        assert g.average(2).shape == (shape[0], shape[1], 1)

    def test_average_weight(self, setup):
        g = setup.g.copy()
        g.grid.fill(1)
        shape = g.shape
        for i in range(3):
            w = np.zeros(shape[i]) + 0.5
            assert (
                g.average(i, weights=w).grid.sum()
                == shape[0] * shape[1] * shape[2] / shape[i]
            )

    def test_interp(self, setup):
        shape = np.array(setup.g.shape, np.int32)
        g = setup.g.interp(shape * 2)
        g1 = g.interp(shape)
        # Sadly the interpolation does not work as it really
        # should...
        # One cannot interp down/up and retrieve the same
        # grid... Perhaps this is ok, but not good... :(
        assert np.allclose(setup.g.grid, g1.grid)

    def test_interp_2d(self, setup):
        shape = np.array(setup.g.shape, np.int32)
        g = setup.g.sum(2)
        g = g.interp([g.shape[0] * 2, g.shape[1] * 2, 1])
        g1 = g.interp([shape[0], shape[1], 1])
        # Sadly the interpolation does not work as it really
        # should...
        # One cannot interp down/up and retrieve the same
        # grid... Perhaps this is ok, but not good... :(
        assert np.allclose(setup.g.sum(2).grid, g1.grid)

    def test_interp_1d(self, setup):
        shape = np.array(setup.g.shape, np.int32)
        g = setup.g.sum(2).sum(0)
        g = g.interp([1, g.shape[1] * 2, 1])
        g1 = g.interp([1, shape[1], 1])
        # Sadly the interpolation does not work as it really
        # should...
        # One cannot interp down/up and retrieve the same
        # grid... Perhaps this is ok, but not good... :(
        assert np.allclose(setup.g.sum(2).sum(0).grid, g1.grid)

    def test_interp_extrap(self, setup):
        shape = np.array(setup.g.shape, np.int32)
        g = setup.g.sum(2)
        g = g.interp([g.shape[0] * 2, g.shape[1] * 2, 10])
        g1 = g.interp([shape[0], shape[1], 1])
        # Sadly the interpolation does not work as it really
        # should...
        # One cannot interp down/up and retrieve the same
        # grid... Perhaps this is ok, but not good... :(
        assert np.allclose(setup.g.sum(2).grid, g1.grid)

    def test_isosurface_orthogonal(self, setup):
        pytest.importorskip("skimage", reason="scikit-image not available")

        # Build an empty grid
        grid = Grid(0.1, lattice=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # Fill it with some values that have a clear isosurface
        grid.grid = np.tile([1, 2, 3, 4, 5, 4, 3, 2, 1, 0], 100).reshape(10, 10, 10)

        # Calculate the isosurface for the value of 2.5
        verts, *returns = grid.isosurface(2.5)

        # The third dimension should contain only two coordinates
        # [1, 2, (HERE) 3, 4, 5, 4, 3, (HERE) 2, 1, 0]
        assert np.unique(verts[:, 1]).shape == (10,)
        assert np.unique(verts[:, 2]).shape == (2,)

    def test_isosurface_non_orthogonal(self, setup):
        pytest.importorskip("skimage", reason="scikit-image not available")

        # If the grid is non-orthogonal, there should be 20 unique values
        # (10 for each (HERE), see test_isosurface_orthogonal)
        grid = Grid(0.1, lattice=[[1, 0, 0], [0, 1, 0], [0, 2, 1]])

        grid.grid = np.tile([1, 2, 3, 4, 5, 4, 3, 2, 1, 0], 100).reshape(10, 10, 10)

        verts, *returns = grid.isosurface(2.5)

        # we have twice as many since the 3rd lattice vector has components in y
        assert np.unique(verts[:, 1]).shape == (20,)
        assert np.unique(verts[:, 2]).shape == (2,)

    def test_smooth_gaussian(self, setup):
        g = Grid(0.1, lattice=[[2, 0, 0], [0, 2, 0], [0, 0, 2]])
        g[10, 10, 10] = 1

        # With a sigma of 0.3 Ang, that single value should be propagated
        # to the whole grid
        smoothed = g.smooth(r=0.7, truncate=4)
        assert not np.any(smoothed.grid == 0)

        # With a sigma of 0.2 Ang, borders should be still 0
        smoothed = g.smooth(r=0.1, truncate=4)
        assert np.any(smoothed.grid == 0)

    def test_smooth_uniform(self, setup):
        g = Grid(0.1, lattice=[[2, 0, 0], [0, 2, 0], [0, 0, 2]])
        g[10, 10, 10] = 1

        # With a radius of 0.7 Ang, that single value should be propagated
        # to the whole grid
        smoothed = g.smooth(r=1.0, method="uniform")
        assert not np.any(smoothed.grid == 0)

        # With a sigma of 0.1 Ang, borders should still be 0
        smoothed = g.smooth(r=0.9, method="uniform")
        assert np.any(smoothed.grid == 0)

    def test_index_ndim1(self, setup):
        mid = np.array(setup.g.shape, np.int32) // 2 - 1
        v = [0.001, 0.0, 0.001]
        idx = setup.g.index(setup.lattice.center() - v)
        assert np.all(mid == idx)
        for i in range(3):
            idx = setup.g.index(setup.lattice.center() - v, axis=i)
            assert idx == mid[i]

    def test_index_fail(self, setup):
        with pytest.raises(ValueError):
            setup.g.index([0.1, 0.2])

    def test_index_ndim2(self, setup):
        mid = np.array(setup.g.shape, np.int32) // 2 - 1
        v = [0.001, 0.0, 0.001]
        idx = setup.g.index([[0] * 3, setup.lattice.center() - v])
        assert np.allclose([[0] * 3, mid], idx)

        for i in range(3):
            idx = setup.g.index([[0] * 3, setup.lattice.center() - v], axis=i)
            assert np.allclose([[0, 0, 0][i], mid[i]], idx)

    def test_index_shape1(self, setup):
        g = setup.g.copy()
        n = 0
        for r in [0.5, 1.0, 1.5]:
            s = Ellipsoid(r)
            idx = g.index(s)
            assert len(idx) > n
            n = len(idx)

        # Check that if we place the sphere an integer
        # amount of cells away we retain an equal amount of indices
        # Also we can check whether they are the same if we add the
        # offset
        v = g.dcell.sum(0)
        vd = v * 0.001
        s = Ellipsoid(1.0)
        idx0 = g.index(s)
        idx0.sort(0)
        for d in [10, 15, 20, 60, 100, 340]:
            idx = g.index(v * d + vd)
            s = Ellipsoid(1.0, center=v * d + vd)
            idx1 = g.index(s)
            idx1.sort(0)
            assert len(idx1) == len(idx0)
            assert np.all(idx0 == idx1 - idx.reshape(1, 3))

    def test_index_shape2(self, setup):
        g = setup.g.copy()
        n = 0
        for r in [0.5, 1.0, 1.5]:
            s = Cuboid(r)
            idx = g.index(s)
            assert len(idx) > n
            n = len(idx)

        # Check that if we place the sphere an integer
        # amount of cells away we retain an equal amount of indices
        # Also we can check whether they are the same if we add the
        # offset
        v = g.dcell.sum(0)
        vd = v * 0.001
        s = Cuboid(1.0)
        idx0 = g.index(s)
        idx0.sort(0)
        for d in [10, 15, 20, 60, 100, 340]:
            idx = g.index(v * d + vd)
            s = Cuboid(1.0, center=v * d + vd)
            idx1 = g.index(s)
            idx1.sort(0)
            assert len(idx1) == len(idx0)
            assert np.all(idx0 == idx1 - idx.reshape(1, 3))

    def test_sum(self, setup):
        for i in range(3):
            assert setup.g.sum(i).shape[i] == 1

    def test_mean(self, setup):
        for i in range(3):
            assert setup.g.mean(i).shape[i] == 1

    def test_cross_section(self, setup):
        for i in range(3):
            assert setup.g.cross_section(1, i).shape[i] == 1

    def test_cross_section_fail(self, setup):
        with pytest.raises(ValueError):
            setup.g.cross_section(1, -1)

    def test_remove_part(self, setup):
        for i in range(3):
            assert setup.g.remove_part(1, i, above=True).shape[i] == 1

    def test_sub_fail(self, setup):
        g = Grid(np.array(setup.g.shape) // 2 + 1, lattice=setup.g.lattice.copy())
        with pytest.raises(ValueError):
            g.sub([], 0)

    def test_sub_part(self, setup):
        for i in range(3):
            assert setup.g.sub_part(1, i, above=False).shape[i] == 1
            assert setup.g.sub_part(1, i, above=True).shape[i] == setup.g.shape[i] - 1

    def test_sub(self, setup):
        for i in range(3):
            assert setup.g.sub(1, i).shape[i] == 1
        for i in range(3):
            assert setup.g.sub([1, 2], i).shape[i] == 2

    def test_remove(self, setup):
        for i in range(3):
            assert setup.g.remove(1, i).shape[i] == setup.g.shape[i] - 1
        for i in range(3):
            assert setup.g.remove([1, 2], i).shape[i] == setup.g.shape[i] - 2

    def test_set_grid1(self, setup):
        g = setup.g.copy()
        g.set_grid([2, 2, 2])
        assert np.all(np.array(g.shape) == 2)
        g.set_grid([2, 2, 3])
        assert np.all(np.array(g.shape) == [2, 2, 3])

    def test_set_grid2(self, setup):
        g = setup.g.copy()
        with pytest.raises(ValueError):
            g.set_grid([2, 2, 2, 4])

    def test_argumentparser(self, setup):
        setup.g.ArgumentParser()

    def test_pyamg1(self, setup):
        g = setup.g.copy()
        g.lattice.set_boundary_condition(g.PERIODIC)  # periodic boundary conditions
        n = np.prod(g.shape)
        A = csr_matrix((n, n))
        b = np.zeros(A.shape[0])

        lb = g.mgrid(slice(0, 1), slice(0, g.shape[1]), slice(0, g.shape[2]))
        lb2 = g.mgrid([slice(0, 1), slice(0, g.shape[1]), slice(0, g.shape[2])])
        assert np.allclose(lb, lb2)
        del lb2

        # Retrieve linear indices
        index = g.pyamg_index(lb)
        g.pyamg_source(b, index, 1)
        assert int(b.sum()) == len(index)
        b[:] = 0

        g.pyamg_fix(A, b, index, 1)
        assert int(b.sum()) == len(index)
        assert A.getnnz() == len(index)

    def test_pyamg2(self, setup):
        # Currently this simply runs the stuff.
        # Nothing is actually tested other than succesfull run,
        # the correctness of the values are not.
        g = setup.g.copy()
        bc = [[g.PERIODIC] * 2, [g.NEUMANN, g.DIRICHLET], [g.DIRICHLET, g.NEUMANN]]
        with pytest.warns(SislWarning, match=r"is having image connections"):
            g.lattice.set_boundary_condition(bc)
        n = np.prod(g.shape)
        A = csr_matrix((n, n))
        b = np.zeros(A.shape[0])
        g.pyamg_boundary_condition(A, b)


def test_grid_fold():
    grid = Grid([4, 5, 6])
    # Assert shapes
    assert grid.index_fold([-1] * 3).shape == (3,)
    assert grid.index_fold([[-1] * 3] * 2, False).shape == (2, 3)
    assert grid.index_fold([[-1] * 3] * 2, True).shape == (1, 3)

    assert np.all(grid.index_fold([-1, -1, -1]) == [3, 4, 5])
    assert np.all(grid.index_fold([[-1, -1, -1]] * 2) == [3, 4, 5])
    assert np.all(grid.index_fold([[-1, -1, -1]] * 2, False) == [[3, 4, 5]] * 2)

    idx = [[-1, 0, 0], [3, 0, 0]]
    assert np.all(grid.index_fold(idx) == [3, 0, 0])
    assert np.all(grid.index_fold(idx, False) == [[3, 0, 0]] * 2)

    idx = [[3, 0, 0], [2, 0, 0]]
    assert np.all(grid.index_fold(idx, False) == idx)
    assert not np.all(grid.index_fold(idx) == idx)  # sorted from unique
    assert np.all(grid.index_fold(idx) == np.sort(idx, axis=0))


def test_grid_tile_sc():
    grid = Grid([4, 5, 6])
    grid2 = grid.tile(2, 2)
    assert grid.shape[:2] == grid2.shape[:2]
    assert grid.shape[2] == grid2.shape[2] // 2
    assert grid.volume * 2 == pytest.approx(grid2.volume)

    grid4 = grid2.tile(2, 1)
    assert grid.shape[0] == grid4.shape[0]
    assert grid.shape[1] == grid4.shape[1] // 2
    assert grid.shape[2] == grid4.shape[2] // 2
    assert grid.volume * 4 == pytest.approx(grid4.volume)


def test_grid_tile_geom():
    grid = Grid([4, 5, 6], geometry=Geometry([0] * 3, Atom[4], lattice=4.0))
    grid2 = grid.tile(2, 2)
    assert grid.shape[:2] == grid2.shape[:2]
    assert grid.shape[2] == grid2.shape[2] // 2
    assert grid.volume * 2 == pytest.approx(grid2.volume)
    assert grid.geometry.na * 2 == grid2.geometry.na

    grid4 = grid2.tile(2, 1)
    assert grid.shape[0] == grid4.shape[0]
    assert grid.shape[1] == grid4.shape[1] // 2
    assert grid.shape[2] == grid4.shape[2] // 2
    assert grid.volume * 4 == pytest.approx(grid4.volume)
    assert grid.geometry.na * 4 == grid4.geometry.na


def test_grid_tile_commensurate():
    gr = geom.graphene()
    gr_lat = gr.lattice.tile(2, 0).tile(2, 1)
    grid = Grid([4, 5, 6], geometry=gr, lattice=gr_lat)
    str(grid)
    grid2 = grid.tile(2, 2)
    assert grid.shape[:2] == grid2.shape[:2]
    assert grid.shape[2] == grid2.shape[2] // 2
    assert grid.volume * 2 == pytest.approx(grid2.volume)
    assert grid.geometry.na * 2 == grid2.geometry.na


def test_grid_tile_in_commensurate():
    gr = geom.graphene()
    lat = Lattice(4.0)
    grid = Grid([4, 5, 6], geometry=gr, lattice=lat)
    str(grid)
    with pytest.raises(SislError):
        grid.tile(2, 2)


def test_new_inherit_class():
    gr = geom.graphene()
    g = Grid([4, 5, 6], geometry=gr)

    class MyClass(Grid):
        pass

    g2 = MyClass.new(g)
    assert g2.__class__ == MyClass
    assert g2 == g
    g2 = Grid.new(g)
    assert g2.__class__ == Grid
    assert g2 == g



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.vasp.doscar import *

pytestmark = [pytest.mark.io, pytest.mark.vasp]


def test_graphene_doscar(sisl_files):
    f = sisl_files("vasp", "graphene", "DOSCAR")
    E, DOS = doscarSileVASP(f).read_data()

    EHa, DOSHa = doscarSileVASP(f).read_data(units=("Ha", "Bohr"))
    assert not np.allclose(E, EHa)
    assert not np.allclose(DOS, DOSHa)
